"""Module with methods to compute the constraint matrix in 1D, 2D, 3D"""

import numpy as np


def constraints_1D(s: float, I: int) -> tuple[np.ndarray, np.ndarray]:
    """Compute the constraints matrix A and the margins vector s in 1D.

    This will define the raking optimization problem:
        min_beta f(beta,y) s.t. A beta = s

    Parameters
    ----------
    s : float
        Target sum of the observations s = sum_i y_i
    I : int
        Number of possible values for categorical variable 1
    Returns
    -------
    A : np.ndarray
        1 * I constraints matrix
    s : np.ndarray
        length 1 margins vector
    """
    assert isinstance(s, float), (
        "The target sum of the observations must be a float."
    )
    assert s >= 0.0, (
        "The target sum of the observations must be positive or null."
    )
    assert isinstance(I, int), (
        "The number of possible values taken by the categorical variable must be an integer."
    )
    assert I > 1, (
        "The number of possible values taken by the categorical variable must be higher than 1."
    )

    A = np.ones((1, I))
    s = np.array([s])
    return (A, s)


def constraints_2D(
    s1: np.ndarray,
    s2: np.ndarray,
    I: int,
    J: int,
    rtol: float = 1e-05,
    atol: float = 1e-08,
) -> tuple[np.ndarray, np.ndarray]:
    """Compute the constraints matrix A and the margins vector s in 2D.

    This will define the raking optimization problem:
        min_beta f(beta,y) s.t. A beta = s

    Parameters
    ----------
    s1 : np.ndarray
        Target sums over rows of the observations table
    s2 : np.ndarray
        Target sums over columns of the observations table
    I : int
        Number of possible values for categorical variable 1
    J : int
        Number of possible values for categorical variable 2
    rtol : float
        Relative tolerance to check whether the margins are consistant. See numpy.allclose documentation for details.
    atol : float
        Absolute tolerance to check whether the margins are consistant. See numpy.allclose documentation for details.

    Returns
    -------
    A : np.ndarray
        (I + J - 1) * (I J) constraints matrix
    s : np.ndarray
        length (I + J) margins vector
    """
    assert isinstance(I, int), (
        "The number of possible values taken by the first categorical variable must be an integer."
    )
    assert I > 1, (
        "The number of possible values taken by the first categorical variable must be higher than 1."
    )
    assert isinstance(J, int), (
        "The number of possible values taken by the second categorical variable must be an integer."
    )
    assert J > 1, (
        "The number of possible values taken by the second categorical variable must be higher than 1."
    )
    assert isinstance(s1, np.ndarray), (
        "The target sums over rows of the observation table must be a Numpy array."
    )
    assert len(s1.shape) == 1, (
        "The target sums over rows of the observation table must be a 1D Numpy array."
    )
    assert isinstance(s2, np.ndarray), (
        "The target sums over columns of the observation table must be a Numpy array."
    )
    assert len(s2.shape) == 1, (
        "The target sums over rows of the observation table must be a 1D Numpy array."
    )
    assert np.all(s1 >= 0.0), (
        "The target sums over rows of the observation table must be positive or null."
    )
    assert np.all(s2 >= 0.0), (
        "The target sums over columns of the observation table must be positive or null."
    )
    assert len(s1) == J, (
        "The target sums over rows must be equal to the number of columns in the observation table."
    )
    assert len(s2) == I, (
        "The target sums over columns must be equal to the number of rows in the observation table."
    )
    assert np.allclose(np.sum(s1), np.sum(s2), rtol, atol), (
        "The sum of the row margins must be equal to the sum of the column margins."
    )

    A = np.zeros((J + I - 1, I * J))
    for j in range(0, J):
        for i in range(0, I - 1):
            A[J + i, j * I + i] = 1
            A[j, j * I + i] = 1
        A[j, j * I + I - 1] = 1
    s = np.concatenate([s1, s2[0 : (I - 1)]])
    return (A, s)


def constraints_3D(
    s1: np.ndarray,
    s2: np.ndarray,
    s3: np.ndarray,
    I: int,
    J: int,
    K: int,
    rtol: float = 1e-05,
    atol: float = 1e-08,
) -> tuple[np.ndarray, np.ndarray]:
    """Compute the constraints matrix A and the margins vector s in 3D.

    This will define the raking optimization problem:
        min_beta f(beta,y) s.t. A beta = s
    The input margins are 3 matrices s1, s2 and s3 and we have:
        sum_i beta_ijk = s1_jk for all j,k
        sum_j beta_ijk = s2_ik for all i,k
        sum_k beta_ijk = s3_ij for all i,j

    Parameters
    ----------
    s1 : np.ndarray
        Target sums over dimension 1 of the observations array
    s2 : np.ndarray
        Target sums over dimension 2 of the observations array
    s3 : np.ndarray
        Target sums over dimension 3 of the observations array
    I : int
        Number of possible values for categorical variable 1
    J : int
        Number of possible values for categorical variable 2
    K : int
        Number of possible values for categorical variable 3
    rtol : float
        Relative tolerance to check whether the margins are consistant. See numpy.allclose documentation for details.
    atol : float
        Absolute tolerance to check whether the margins are consistant. See numpy.allclose documentation for details.

    Returns
    -------
    A : np.ndarray
        (I J + I K + J K - I - J - K + 1) * (I J K) constraints matrix
    s : np.ndarray
        length (I J + I K + J K - I - J - K + 1) margins vector
    """
    assert isinstance(I, int), (
        "The number of possible values taken by the first categorical variable must be an integer."
    )
    assert I > 1, (
        "The number of possible values taken by the first categorical variable must be higher than 1."
    )
    assert isinstance(J, int), (
        "The number of possible values taken by the second categorical variable must be an integer."
    )
    assert J > 1, (
        "The number of possible values taken by the second categorical variable must be higher than 1."
    )
    assert isinstance(K, int), (
        "The number of possible values taken by the third categorical variable must be an integer."
    )
    assert K > 1, (
        "The number of possible values taken by the third categorical variable must be higher than 1."
    )

    assert isinstance(s1, np.ndarray), (
        "The target sums over dimension 1 of the observation array must be a Numpy array."
    )
    assert len(s1.shape) == 2, (
        "The target sums over dimension 1 of the observation array must be a 2D Numpy array."
    )
    assert s1.shape[0] == J, (
        "The target sums over dimension 1 must have {} rows.".format(J)
    )
    assert s1.shape[1] == K, (
        "The target sums over dimension 1 must have {} columns.".format(K)
    )

    assert isinstance(s2, np.ndarray), (
        "The target sums over dimension 2 of the observation array must be a Numpy array."
    )
    assert len(s2.shape) == 2, (
        "The target sums over dimension 2 of the observation array must be a 2D Numpy array."
    )
    assert s2.shape[0] == I, (
        "The target sums over dimension 2 must have {} rows.".format(I)
    )
    assert s2.shape[1] == K, (
        "The target sums over dimension 2 must have {} columns.".format(K)
    )

    assert isinstance(s3, np.ndarray), (
        "The target sums over dimension 3 of the observation array must be a Numpy array."
    )
    assert len(s3.shape) == 2, (
        "The target sums over dimension 3 of the observation array must be a 2D Numpy array."
    )
    assert s3.shape[0] == I, (
        "The target sums over dimension 3 must have {} rows.".format(I)
    )
    assert s3.shape[1] == J, (
        "The target sums over dimension 3 must have {} columns.".format(J)
    )

    assert np.all(s1 >= 0.0), (
        "The target sums over dimension 1 of the observation array must be positive or null."
    )
    assert np.all(s2 >= 0.0), (
        "The target sums over dimension 2 of the observation array must be positive or null."
    )
    assert np.all(s3 >= 0.0), (
        "The target sums over dimension 3 of the observation array must be positive or null."
    )

    assert np.allclose(np.sum(s1, axis=0), np.sum(s2, axis=0), rtol, atol), (
        "The sums of the targets for dimension 1 and 2 must be equal."
    )
    assert np.allclose(np.sum(s2, axis=1), np.sum(s3, axis=1), rtol, atol), (
        "The sums of the targets for dimension 2 and 3 must be equal."
    )
    assert np.allclose(np.sum(s1, axis=1), np.sum(s3, axis=0), rtol, atol), (
        "The sums of the targets for dimension 1 and 3 must be equal."
    )

    A = np.zeros((I * J + I * K + J * K - I - J - K + 1, I * J * K))
    s = np.zeros(I * J + I * K + J * K - I - J - K + 1)
    for k in range(0, K):
        for j in range(0, J - 1):
            for i in range(0, I):
                A[(J - 1) * k + j, I * J * k + I * j + i] = 1
            s[(J - 1) * k + j] = s1[j, k]
    for i in range(0, I):
        A[(J - 1) * K, I * J * (K - 1) + I * (J - 1) + i] = 1
    s[(J - 1) * K] = s1[J - 1, K - 1]
    for i in range(0, I):
        for k in range(0, K - 1):
            for j in range(0, J):
                A[(J - 1) * K + 1 + (K - 1) * i + k, I * J * k + I * j + i] = 1
            s[(J - 1) * K + 1 + (K - 1) * i + k] = s2[i, k]
    for j in range(0, J):
        for i in range(0, I - 1):
            for k in range(0, K):
                A[
                    (J - 1) * K + 1 + (K - 1) * I + (I - 1) * j + i,
                    I * J * k + I * j + i,
                ] = 1
            s[(J - 1) * K + 1 + (K - 1) * I + (I - 1) * j + i] = s3[i, j]
    return (A, s)


def constraints_USHD(
    s_cause: np.ndarray,
    I: int,
    J: int,
    K: int,
    rtol: float = 1e-05,
    atol: float = 1e-08,
) -> tuple[np.ndarray, np.ndarray]:
    """Compute the constraints matrix A and the margins vector s for the USHD use case.

    This will define the raking optimization problem:
        min_beta f(beta,y) s.t. A beta = s
    The input margins are the 1 + I values:
        - beta_000 = Total number of deaths (all causes, all races, at the state level)
        - beta_i00 = Number of deaths for cause i (all races, at the state level)

    Parameters
    ----------
    s_cause : np.ndarray
        Total number of deaths (all causes, and each cause)
    I : int
        Number of causes of deaths
    J : int
        Number of races and ethnicities
    K : int
        Number of counties
    rtol : float
        Relative tolerance to check whether the margins are consistant. See numpy.allclose documentation for details.
    atol : float
        Absolute tolerance to check whether the margins are consistant. See numpy.allclose documentation for details.

    Returns
    -------
    A : np.ndarray
        (I + 2 * K + J * K + (I - 1) * K) * ((I + 1) * (J + 1) * K) constraints matrix
    s : np.ndarray
        length (I + 2 * K + J * K + (I - 1) * K) margins vector
    """
    assert isinstance(I, int), (
        "The number of causes of deaths must be an integer."
    )
    assert I > 1, "The number of causes of deaths must be higher than 1."
    assert isinstance(J, int), (
        "The number of races and ethnicities must be an integer."
    )
    assert J > 1, "The number of races and ethnicities must be higher than 1."
    assert isinstance(K, int), "The number of counties must be an integer."
    assert K > 1, "The number of counties must be higher than 1."

    assert isinstance(s_cause, np.ndarray), (
        "The margins vector for the causes of death must be a Numpy array."
    )
    assert len(s_cause.shape) == 1, (
        "The margins vector for the causes of death must be a 1D Numpy array."
    )
    assert np.all(s_cause >= 0.0), (
        "The number of deaths for each cause must be positive or null."
    )
    assert len(s_cause) == I + 1, (
        "The length of the margins vector for the causes of death must be equal to 1 + number of causes."
    )

    assert np.allclose(s_cause[0], np.sum(s_cause[1:]), rtol, atol), (
        "The all-causes number of deaths must be equal to the sum of the numbers of deaths per cause."
    )

    A = np.zeros((I + 2 * K + J * K + (I - 1) * K, (I + 1) * (J + 1) * K))
    s = np.zeros(I + 2 * K + J * K + (I - 1) * K)
    # Constraint sum_k=0,...,K-1 beta_i,0,k = s_i for i=1,...,I
    for i in range(0, I):
        for k in range(0, K):
            A[i, k * (I + 1) * (J + 1) + i + 1] = 1
        s[i] = s_cause[i + 1]
    # Constraint sum_i=1,...,I beta_i,0,k - beta_0,0,k = 0 for k=0,...,K-1
    for k in range(0, K):
        for i in range(1, I + 1):
            A[I + k, k * (I + 1) * (J + 1) + i] = 1
        A[I + k, k * (I + 1) * (J + 1)] = -1
    # Constraint sum_j=1,...,J beta_0,j,k - beta_0,0,k = 0 for k=0,...,K-1
    for k in range(0, K):
        for j in range(1, J + 1):
            A[I + K + k, k * (I + 1) * (J + 1) + j * (I + 1)] = 1
        A[I + K + k, k * (I + 1) * (J + 1)] = -1
    # Constraint sum_i=1,...,I beta_i,j,k - beta_0,j,k = 0 for j=1,...,J and k=0,...,K-1
    for k in range(0, K):
        for j in range(1, J + 1):
            for i in range(1, I + 1):
                A[
                    I + 2 * K + k * J + j - 1,
                    k * (I + 1) * (J + 1) + j * (I + 1) + i,
                ] = 1
            A[
                I + 2 * K + k * J + j - 1, k * (I + 1) * (J + 1) + j * (I + 1)
            ] = -1
    # Constraint sum_j=1,...,J beta_i,j,k - beta_i,0,k = 0 for i=1,...,I and k=0,...,K-1
    for k in range(0, K):
        for i in range(1, I):
            for j in range(1, J + 1):
                A[
                    I + 2 * K + J * K + k * (I - 1) + i - 1,
                    k * (I + 1) * (J + 1) + j * (I + 1) + i,
                ] = 1
            A[
                I + 2 * K + J * K + k * (I - 1) + i - 1,
                k * (I + 1) * (J + 1) + i,
            ] = -1
    return (A, s)



"""Module with methods to format the data for the raking methods"""

import numpy as np
import pandas as pd

pd.options.mode.chained_assignment = None


def format_data_1D(
    df_obs: pd.DataFrame,
    df_margins: pd.DataFrame,
    var_name: str,
    weights: str = None,
    lower: str = None,
    upper: str = None,
) -> tuple[
    np.ndarray,
    float,
    int,
    np.ndarray | None,
    np.ndarray | None,
    np.ndarray | None,
]:
    """Read the data and create the inputs of the raking functions (1D problem).

    Parameters
    ----------
    df_obs : pd.DataFrame
        Observations data
    df_margins : pd.DataFrame
        Margins data
    var_name : string
        Name of the variable over which we rake (e.g. cause, race, county)
    weights : string
        Name of the column containing the raking weights
    lower : string
        Name of the column containing the lower boundaries (for logit raking)
    upper : string
        Name of the column containing the upper boundaries (for logit raking)

    Returns
    -------
    y : np.ndarray
        Vector of observations
    s : float
        Target sum of the observations s = sum_i y_i
    I : int
        Number of possible values for categorical variable 1
    q : np.ndarray
        Vector of weights
    l : np.ndarray
        Lower bounds for the observations
    h : np.ndarray
        Upper bounds for the observations
    """
    assert isinstance(df_obs, pd.DataFrame), (
        "The observations should be a pandas data frame."
    )
    assert len(df_obs) >= 2, (
        "There should be at least 2 data points for the observations."
    )

    assert isinstance(df_margins, pd.DataFrame), (
        "The margins should be a pandas data frame."
    )
    assert len(df_margins) == 1, (
        "There should be only one data point for the margins."
    )

    assert "value" in df_obs.columns.tolist(), (
        "The observations data frame should contain a value column."
    )
    assert isinstance(var_name, str), (
        "The name of the categorical variable should be a string."
    )
    assert var_name in df_obs.columns.tolist(), (
        "The column for the categorical variable "
        + var_name
        + " is missing from the observations data frame."
    )

    assert "value_agg_over_" + var_name in df_margins.columns.tolist(), (
        "The column for the aggregated value over "
        + var_name
        + " is missing from the margins data frame."
    )

    if weights is not None:
        assert isinstance(weights, str), (
            "The name of the column containing the weights should be a string."
        )
        assert weights in df_obs.columns.tolist(), (
            "The column containing the weights is missing from the data frame."
        )
    if lower is not None:
        assert isinstance(lower, str), (
            "The name of the column containing the lower boundaries should be a string."
        )
        assert lower in df_obs.columns.tolist(), (
            "The column containing the lower boundaries is missing from the data frame."
        )
    if upper is not None:
        assert isinstance(upper, str), (
            "The name of the column containing the upper boundaries should be a string."
        )
        assert upper in df_obs.columns.tolist(), (
            "The column containing the upper_boundaries is missing from the data frame."
        )

    # Check the observations data
    assert df_obs.value.isna().sum() == 0, (
        "There are missing values in the value column of the observations."
    )
    assert df_obs[var_name].isna().sum() == 0, (
        "There are missing values in the "
        + var_name
        + " column of the observations."
    )
    assert len(df_obs[df_obs.duplicated([var_name])]) == 0, (
        "There are duplicated rows in the observations."
    )

    # Check the margins data
    assert df_margins["value_agg_over_" + var_name].isna().sum() == 0, (
        "There are missing values in the value_agg_over"
        + var_name
        + " column of the margins."
    )

    # Create input variables for the raking functions
    df_obs.sort_values(by=[var_name], inplace=True)
    I = len(df_obs[var_name].unique())
    y = df_obs.value.to_numpy()
    s = df_margins["value_agg_over_" + var_name].to_numpy()[0]
    if weights is not None:
        q = df_obs[weights].to_numpy()
    else:
        q = None
    if lower is not None:
        l = df_obs[lower].to_numpy()
    else:
        l = None
    if upper is not None:
        h = df_obs[upper].to_numpy()
    else:
        h = None
    return (y, s, I, q, l, h)


def format_data_2D(
    df_obs: pd.DataFrame,
    df_margins_1: pd.DataFrame,
    df_margins_2: pd.DataFrame,
    var_names: list,
    weights: str = None,
    lower: str = None,
    upper: str = None,
) -> tuple[
    np.ndarray,
    np.ndarray,
    np.ndarray,
    int,
    int,
    np.ndarray | None,
    np.ndarray | None,
    np.ndarray | None,
]:
    """Read the data and create the inputs of the raking functions (2D problem).

    Parameters
    ----------
    df_obs : pd.DataFrame
        Observations data
    df_margins_1 : pd.DataFrame
        Margins data (sums over the first variable)
    df_margins_2 : pd.DataFrame
        Margins data (sums over the second variable)
    var_names : list of 2 strings
        Names of the two variables over which we rake (e.g. cause, race, county)
    weights : string
        Name of the column containing the raking weights
    lower : string
        Name of the column containing the lower boundaries (for logit raking)
    upper : string
        Name of the column containing the upper boundaries (for logit raking)

    Returns
    -------
    y : np.ndarray
        Vector of observations
    s1 : np.ndarray
        Target sums of the observations over categorical variable 1
    s2 : np.ndarray
        Target sums of the observations over categorical variable 2
    I : int
        Number of possible values for categorical variable 1
    J : int
        Number of possible values for categorical variable 2
    q : np.ndarray
        Vector of weights
    l : np.ndarray
        Lower bounds for the observations
    h : np.ndarray
        Upper bounds for the observations
    """
    assert isinstance(df_obs, pd.DataFrame), (
        "The observations should be a pandas data frame."
    )
    assert len(df_obs) >= 4, (
        "There should be at least 4 data points for the observations."
    )

    assert isinstance(df_margins_1, pd.DataFrame), (
        "The margins for the first variable should be a pandas data frame."
    )
    assert len(df_margins_1) >= 2, (
        "There should be at least 2 data points for the first margins."
    )

    assert isinstance(df_margins_2, pd.DataFrame), (
        "The margins for the second variable should be a pandas data frame."
    )
    assert len(df_margins_2) >= 2, (
        "There should be at least 2 data points for the second margins."
    )

    assert "value" in df_obs.columns.tolist(), (
        "The observations data frame should contain a value column."
    )
    assert isinstance(var_names, list), (
        "Please enter the names of the columns containing the values of the categorical variables as a list."
    )
    assert len(var_names) == 2, "You should have 2 categorical variables."
    for var_name in var_names:
        assert isinstance(var_name, str), (
            "The name of the categorical variable "
            + str(var_name)
            + " should be a string."
        )
        assert var_name in df_obs.columns.tolist(), (
            "The column for the categorical variable "
            + var_name
            + " is missing from the observations data frame."
        )

    assert var_names[1] in df_margins_1.columns.tolist(), (
        "The column for the categorigal variable "
        + var_name[1]
        + " is missing from the first margins data frame."
    )
    assert "value_agg_over_" + var_names[0] in df_margins_1.columns.tolist(), (
        "The column for the aggregated value over "
        + var_names[0]
        + " is missing from the first margins data frame."
    )

    assert var_names[0] in df_margins_2.columns.tolist(), (
        "The column for the categorigal variable "
        + var_name[0]
        + " is missing from the second margins data frame."
    )
    assert "value_agg_over_" + var_names[1] in df_margins_2.columns.tolist(), (
        "The column for the aggregated value over "
        + var_names[1]
        + " is missing from the second margins data frame."
    )

    if weights is not None:
        assert isinstance(weights, str), (
            "The name of the column containing the weights should be a string."
        )
        assert weights in df_obs.columns.tolist(), (
            "The column containing the weights is missing from the data frame."
        )
    if lower is not None:
        assert isinstance(lower, str), (
            "The name of the column containing the lower boundaries should be a string."
        )
        assert lower in df_obs.columns.tolist(), (
            "The column containing the lower boundaries is missing from the data frame."
        )
    if upper is not None:
        assert isinstance(upper, str), (
            "The name of the column containing the upper boundaries should be a string."
        )
        assert upper in df_obs.columns.tolist(), (
            "The column containing the upper_boundaries is missing from the data frame."
        )

    # Check the observations data
    assert df_obs.value.isna().sum() == 0, (
        "There are missing values in the value column of the observations."
    )
    for var_name in var_names:
        assert df_obs[var_name].isna().sum() == 0, (
            "There are missing values in the "
            + var_name
            + " column of the observations."
        )
    assert len(df_obs[df_obs.duplicated(var_names)]) == 0, (
        "There are duplicated rows in the observations."
    )
    count_obs = df_obs[var_names].value_counts()
    assert (len(count_obs.unique()) == 1) and (count_obs.unique()[0] == 1), (
        "There are missing combinations of "
        + var_names[0]
        + " and "
        + var_names[1]
        + " in the observations."
    )

    # Check the first margins data
    assert df_margins_1[var_names[1]].isna().sum() == 0, (
        "There are missing values in the "
        + var_names[1]
        + " column of the margins."
    )
    assert df_margins_1["value_agg_over_" + var_names[0]].isna().sum() == 0, (
        "There are missing values in the value_agg_over"
        + var_names[0]
        + " column of the margins."
    )
    assert len(df_margins_1[df_margins_1.duplicated([var_names[1]])]) == 0, (
        "There are duplicated rows in the first margins data frame."
    )

    # Check the second margins data
    assert df_margins_2[var_names[0]].isna().sum() == 0, (
        "There are missing values in the "
        + var_names[0]
        + " column of the margins."
    )
    assert df_margins_2["value_agg_over_" + var_names[1]].isna().sum() == 0, (
        "There are missing values in the value_agg_over"
        + var_names[1]
        + " column of the margins."
    )
    assert len(df_margins_2[df_margins_2.duplicated([var_names[0]])]) == 0, (
        "There are duplicated rows in the second margins data frame."
    )

    # Check consistency between observations and margins
    assert len(df_obs[var_names[0]].unique()) == len(
        df_margins_2[var_names[0]].unique()
    ), (
        "The number of categories for "
        + var_names[0]
        + " should be the same in the observations and margins data frames."
    )
    assert set(df_obs[var_names[0]].unique().tolist()) == set(
        df_margins_2[var_names[0]].unique().tolist()
    ), (
        "The names of the categories for "
        + var_names[0]
        + " should be the same in the observations and margins data frames."
    )
    assert len(df_obs[var_names[1]].unique()) == len(
        df_margins_1[var_names[1]].unique()
    ), (
        "The number of categories for "
        + var_names[1]
        + " should be the same in the observations and margins data frames."
    )
    assert set(df_obs[var_names[1]].unique().tolist()) == set(
        df_margins_1[var_names[1]].unique().tolist()
    ), (
        "The names of the categories for "
        + var_names[1]
        + " should be the same in the observations and margins data frames."
    )

    # Create input variables for the raking functions
    df_obs.sort_values(by=[var_names[1], var_names[0]], inplace=True)
    df_margins_1.sort_values(by=[var_names[1]], inplace=True)
    df_margins_2.sort_values(by=[var_names[0]], inplace=True)
    I = len(df_obs[var_names[0]].unique())
    J = len(df_obs[var_names[1]].unique())
    y = df_obs.value.to_numpy()
    s1 = df_margins_1["value_agg_over_" + var_names[0]].to_numpy()
    s2 = df_margins_2["value_agg_over_" + var_names[1]].to_numpy()
    if weights is not None:
        q = df_obs[weights].to_numpy()
    else:
        q = None
    if lower is not None:
        l = df_obs[lower].to_numpy()
    else:
        l = None
    if upper is not None:
        h = df_obs[upper].to_numpy()
    else:
        h = None
    return (y, s1, s2, I, J, q, l, h)


def format_data_3D(
    df_obs: pd.DataFrame,
    df_margins_1: pd.DataFrame,
    df_margins_2: pd.DataFrame,
    df_margins_3: pd.DataFrame,
    var_names: list,
    weights: str = None,
    lower: str = None,
    upper: str = None,
) -> tuple[
    np.ndarray,
    np.ndarray,
    np.ndarray,
    np.ndarray,
    int,
    int,
    int,
    np.ndarray | None,
    np.ndarray | None,
    np.ndarray | None,
]:
    """Read the data and create the inputs of the raking functions (3D problem).

    Parameters
    ----------
    df_obs : pd.DataFrame
        Observations data
    df_margins_1 : pd.DataFrame
        Margins data (sums over the first variable)
    df_margins_2 : pd.DataFrame
        Margins data (sums over the second variable)
    df_margins_3 : pd.DataFrame
        Margins data (sums over the third variable)
    var_names : list of 3 strings
        Names of the three variables over which we rake (e.g. cause, race, county)
    weights : string
        Name of the column containing the raking weights
    lower : string
        Name of the column containing the lower boundaries (for logit raking)
    upper : string
        Name of the column containing the upper boundaries (for logit raking)

    Returns
    -------
    y : np.ndarray
        Vector of observations
    s1 : np.ndarray
        Target sums of the observations over categorical variable 1
    s2 : np.ndarray
        Target sums of the observations over categorical variable 2
    s3 : np.ndarray
        Target sums of the observations over categorical variable 3
    I : int
        Number of possible values for categorical variable 1
    J : int
        Number of possible values for categorical variable 2
    K : int
        Number of possible values for categorical variable 3
    q : np.ndarray
        Vector of weights
    l : np.ndarray
        Lower bounds for the observations
    h : np.ndarray
        Upper bounds for the observations
    """
    assert isinstance(df_obs, pd.DataFrame), (
        "The observations should be a pandas data frame."
    )
    assert len(df_obs) >= 8, (
        "There should be at least 8 data points for the observations."
    )

    assert isinstance(df_margins_1, pd.DataFrame), (
        "The margins for the first variable should be a pandas data frame."
    )
    assert len(df_margins_1) >= 4, (
        "There should be at least 4 data points for the first margins."
    )

    assert isinstance(df_margins_2, pd.DataFrame), (
        "The margins for the second variable should be a pandas data frame."
    )
    assert len(df_margins_2) >= 4, (
        "There should be at least 4 data points for the second margins."
    )

    assert isinstance(df_margins_3, pd.DataFrame), (
        "The margins for the third variable should be a pandas data frame."
    )
    assert len(df_margins_3) >= 4, (
        "There should be at least 4 data points for the third margins."
    )

    assert "value" in df_obs.columns.tolist(), (
        "The observations data frame should contain a value column."
    )
    assert isinstance(var_names, list), (
        "Please enter the names of the columns containing the values of the categorical variables as a list."
    )
    assert len(var_names) == 3, "You should have 3 categorical variables."
    for var_name in var_names:
        assert isinstance(var_name, str), (
            "The name of the categorical variable "
            + +str(var_name)
            + " should be a string."
        )
        assert var_name in df_obs.columns.tolist(), (
            "The column for the categorical variable "
            + var_name
            + " is missing from the observations data frame."
        )

    assert var_names[1] in df_margins_1.columns.tolist(), (
        "The column for the categorigal variable "
        + var_name[1]
        + " is missing from the first margins data frame."
    )
    assert var_names[2] in df_margins_1.columns.tolist(), (
        "The column for the categorigal variable "
        + var_name[2]
        + " is missing from the first margins data frame."
    )
    assert "value_agg_over_" + var_names[0] in df_margins_1.columns.tolist(), (
        "The column for the aggregated value over "
        + var_names[0]
        + " is missing from the first margins data frame."
    )

    assert var_names[0] in df_margins_2.columns.tolist(), (
        "The column for the categorigal variable "
        + var_name[0]
        + " is missing from the second margins data frame."
    )
    assert var_names[2] in df_margins_2.columns.tolist(), (
        "The column for the categorigal variable "
        + var_name[2]
        + " is missing from the second margins data frame."
    )
    assert "value_agg_over_" + var_names[1] in df_margins_2.columns.tolist(), (
        "The column for the aggregated value over "
        + var_names[1]
        + " is missing from the second margins data frame."
    )

    assert var_names[0] in df_margins_3.columns.tolist(), (
        "The column for the categorigal variable "
        + var_name[0]
        + " is missing from the third margins data frame."
    )
    assert var_names[1] in df_margins_3.columns.tolist(), (
        "The column for the categorigal variable "
        + var_name[1]
        + " is missing from the third margins data frame."
    )
    assert "value_agg_over_" + var_names[2] in df_margins_3.columns.tolist(), (
        "The column for the aggregated value over "
        + var_names[2]
        + " is missing from the third margins data frame."
    )

    if weights is not None:
        assert isinstance(weights, str), (
            "The name of the column containing the weights should be a string."
        )
        assert weights in df_obs.columns.tolist(), (
            "The column containing the weights is missing from the data frame."
        )
    if lower is not None:
        assert isinstance(lower, str), (
            "The name of the column containing the lower boundaries should be a string."
        )
        assert lower in df_obs.columns.tolist(), (
            "The column containing the lower boundaries is missing from the data frame."
        )
    if upper is not None:
        assert isinstance(upper, str), (
            "The name of the column containing the upper boundaries should be a string."
        )
        assert upper in df_obs.columns.tolist(), (
            "The column containing the upper_boundaries is missing from the data frame."
        )

    # Check the observations data
    assert df_obs.value.isna().sum() == 0, (
        "There are missing values in the value column of the observations."
    )
    for var_name in var_names:
        assert df_obs[var_name].isna().sum() == 0, (
            "There are missing values in the "
            + var_name
            + " column of the observations."
        )
    assert len(df_obs[df_obs.duplicated(var_names)]) == 0, (
        "There are duplicated rows in the observations."
    )
    count_obs = df_obs[var_names].value_counts()
    assert (len(count_obs.unique()) == 1) and (count_obs.unique()[0] == 1), (
        "There are missing combinations of "
        + var_names[0]
        + ", "
        + var_names[1]
        + " and "
        + var_names[2]
        + " in the observations."
    )

    # Check the first margins data
    assert df_margins_1[var_names[1]].isna().sum() == 0, (
        "There are missing values in the "
        + var_names[1]
        + " column of the first margins."
    )
    assert df_margins_1[var_names[2]].isna().sum() == 0, (
        "There are missing values in the "
        + var_names[2]
        + " column of the first margins."
    )
    assert df_margins_1["value_agg_over_" + var_names[0]].isna().sum() == 0, (
        "There are missing values in the value_agg_over"
        + var_names[0]
        + " column of the first margins."
    )
    assert (
        len(df_margins_1[df_margins_1.duplicated([var_names[1], var_names[2]])])
        == 0
    ), "There are duplicated rows in the first margins data frame."
    count_obs = df_margins_1[[var_names[1], var_names[2]]].value_counts()
    assert (len(count_obs.unique()) == 1) and (count_obs.unique()[0] == 1), (
        "There are missing combinations of "
        + var_names[1]
        + " and "
        + var_names[2]
        + " in the first margins."
    )

    # Check the second margins data
    assert df_margins_2[var_names[0]].isna().sum() == 0, (
        "There are missing values in the "
        + var_names[0]
        + " column of the second margins."
    )
    assert df_margins_2[var_names[2]].isna().sum() == 0, (
        "There are missing values in the "
        + var_names[2]
        + " column of the second margins."
    )
    assert df_margins_2["value_agg_over_" + var_names[1]].isna().sum() == 0, (
        "There are missing values in the value_agg_over"
        + var_names[1]
        + " column of the second margins."
    )
    assert (
        len(df_margins_2[df_margins_2.duplicated([var_names[0], var_names[2]])])
        == 0
    ), "There are duplicated rows in the second margins data frame."
    count_obs = df_margins_2[[var_names[0], var_names[2]]].value_counts()
    assert (len(count_obs.unique()) == 1) and (count_obs.unique()[0] == 1), (
        "There are missing combinations of "
        + var_names[0]
        + " and "
        + var_names[2]
        + " in the second margins."
    )

    # Check the third margins data
    assert df_margins_3[var_names[0]].isna().sum() == 0, (
        "There are missing values in the "
        + var_names[0]
        + " column of the third margins."
    )
    assert df_margins_3[var_names[1]].isna().sum() == 0, (
        "There are missing values in the "
        + var_names[1]
        + " column of the third margins."
    )
    assert df_margins_3["value_agg_over_" + var_names[2]].isna().sum() == 0, (
        "There are missing values in the value_agg_over"
        + var_names[2]
        + " column of the third margins."
    )
    assert (
        len(df_margins_3[df_margins_3.duplicated([var_names[0], var_names[1]])])
        == 0
    ), "There are duplicated rows in the third margins data frame."
    count_obs = df_margins_3[[var_names[0], var_names[1]]].value_counts()
    assert (len(count_obs.unique()) == 1) and (count_obs.unique()[0] == 1), (
        "There are missing combinations of "
        + var_names[0]
        + " and "
        + var_names[1]
        + " in the third margins."
    )

    # Check consistency between observations and first margins
    assert len(df_obs[var_names[1]].unique()) == len(
        df_margins_1[var_names[1]].unique()
    ), (
        "The number of categories for "
        + var_names[1]
        + " should be the same in the observations and first margins data frames."
    )
    assert set(df_obs[var_names[1]].unique().tolist()) == set(
        df_margins_1[var_names[1]].unique().tolist()
    ), (
        "The names of the categories for "
        + var_names[1]
        + " should be the same in the observations and first margins data frames."
    )
    assert len(df_obs[var_names[2]].unique()) == len(
        df_margins_1[var_names[2]].unique()
    ), (
        "The number of categories for "
        + var_names[2]
        + " should be the same in the observations and first margins data frames."
    )
    assert set(df_obs[var_names[2]].unique().tolist()) == set(
        df_margins_1[var_names[2]].unique().tolist()
    ), (
        "The names of the categories for "
        + var_names[2]
        + " should be the same in the observations and first margins data frames."
    )

    # Check consistency between observations and second margins
    assert len(df_obs[var_names[0]].unique()) == len(
        df_margins_2[var_names[0]].unique()
    ), (
        "The number of categories for "
        + var_names[0]
        + " should be the same in the observations and second margins data frames."
    )
    assert set(df_obs[var_names[0]].unique().tolist()) == set(
        df_margins_2[var_names[0]].unique().tolist()
    ), (
        "The names of the categories for "
        + var_names[0]
        + " should be the same in the observations and second margins data frames."
    )
    assert len(df_obs[var_names[2]].unique()) == len(
        df_margins_2[var_names[2]].unique()
    ), (
        "The number of categories for "
        + var_names[2]
        + " should be the same in the observations and second margins data frames."
    )
    assert set(df_obs[var_names[2]].unique().tolist()) == set(
        df_margins_2[var_names[2]].unique().tolist()
    ), (
        "The names of the categories for "
        + var_names[2]
        + " should be the same in the observations and second margins data frames."
    )

    # Check consistency between observations and third margins
    assert len(df_obs[var_names[0]].unique()) == len(
        df_margins_3[var_names[0]].unique()
    ), (
        "The number of categories for "
        + var_names[0]
        + " should be the same in the observations and third margins data frames."
    )
    assert set(df_obs[var_names[0]].unique().tolist()) == set(
        df_margins_3[var_names[0]].unique().tolist()
    ), (
        "The names of the categories for "
        + var_names[0]
        + " should be the same in the observations and third margins data frames."
    )
    assert len(df_obs[var_names[1]].unique()) == len(
        df_margins_3[var_names[1]].unique()
    ), (
        "The number of categories for "
        + var_names[1]
        + " should be the same in the observations and third margins data frames."
    )
    assert set(df_obs[var_names[1]].unique().tolist()) == set(
        df_margins_3[var_names[1]].unique().tolist()
    ), (
        "The names of the categories for "
        + var_names[1]
        + " should be the same in the observations and third margins data frames."
    )

    # Create input variables for the raking functions
    df_obs.sort_values(
        by=[var_names[2], var_names[1], var_names[0]], inplace=True
    )
    df_margins_1.sort_values(by=[var_names[2], var_names[1]], inplace=True)
    df_margins_2.sort_values(by=[var_names[2], var_names[0]], inplace=True)
    df_margins_3.sort_values(by=[var_names[1], var_names[0]], inplace=True)
    I = len(df_obs[var_names[0]].unique())
    J = len(df_obs[var_names[1]].unique())
    K = len(df_obs[var_names[2]].unique())
    y = df_obs.value.to_numpy()
    s1 = (
        df_margins_1["value_agg_over_" + var_names[0]]
        .to_numpy()
        .reshape([J, K], order="F")
    )
    s2 = (
        df_margins_2["value_agg_over_" + var_names[1]]
        .to_numpy()
        .reshape([I, K], order="F")
    )
    s3 = (
        df_margins_3["value_agg_over_" + var_names[2]]
        .to_numpy()
        .reshape([I, J], order="F")
    )
    if weights is not None:
        q = df_obs[weights].to_numpy()
    else:
        q = None
    if lower is not None:
        l = df_obs[lower].to_numpy()
    else:
        l = None
    if upper is not None:
        h = df_obs[upper].to_numpy()
    else:
        h = None
    return (y, s1, s2, s3, I, J, K, q, l, h)


def format_data_USHD(
    df_obs: pd.DataFrame,
    df_margins: pd.DataFrame,
    weights: str = None,
    lower: str = None,
    upper: str = None,
) -> tuple[
    np.ndarray,
    np.ndarray,
    int,
    int,
    int,
    np.ndarray | None,
    np.ndarray | None,
    np.ndarray | None,
]:
    """Read the data and create the inputs of the raking functions (USHD problem).

    Parameters
    ----------
    df_obs : pd.DataFrame
        Observations data
    df_margins : pd.DataFrame
        Margins data (GBD)
    weights : string
        Name of the column containing the raking weights
    lower : string
        Name of the column containing the lower boundaries (for logit raking)
    upper : string
        Name of the column containing the upper boundaries (for logit raking)

    Returns
    -------
    y : np.ndarray
        Vector of observations
    s : np.ndarray
        Total number of deaths (all causes, and each cause)
    I : int
        Number of possible values for cause
    J : int
        Number of possible values for race
    K : int
        Number of possible values for county
    q : np.ndarray
        Vector of weights
    l : np.ndarray
        Lower bounds for the observations
    h : np.ndarray
        Upper bounds for the observations
    """
    assert isinstance(df_obs, pd.DataFrame), (
        "The observations should be a pandas data frame."
    )
    assert len(df_obs) >= 18, (
        "There should be at least 18 data points for the observations."
    )

    assert isinstance(df_margins, pd.DataFrame), (
        "The margins should be a pandas data frame."
    )
    assert len(df_margins) >= 3, (
        "There should be at least 3 data points for the margins."
    )

    for var_name in ["value", "cause", "race", "county"]:
        assert var_name in df_obs.columns.tolist(), (
            "The column for the categorical variable "
            + var_name
            + " is missing from the observations data frame."
        )

    assert "cause" in df_margins.columns.tolist(), (
        "The cause column is missing from the margins data frame."
    )
    assert "value_agg_over_race_county" in df_margins.columns.tolist(), (
        "The column for the aggregated value over races and counties is missing from the margins data frame."
    )

    if weights is not None:
        assert isinstance(weights, str), (
            "The name of the column containing the weights should be a string."
        )
        assert weights in df_obs.columns.tolist(), (
            "The column containing the weights is missing from the data frame."
        )
    if lower is not None:
        assert isinstance(lower, str), (
            "The name of the column containing the lower boundaries should be a string."
        )
        assert lower in df_obs.columns.tolist(), (
            "The column containing the lower boundaries is missing from the data frame."
        )
    if upper is not None:
        assert isinstance(upper, str), (
            "The name of the column containing the upper boundaries should be a string."
        )
        assert upper in df_obs.columns.tolist(), (
            "The column containing the upper_boundaries is missing from the data frame."
        )

    # Check the observations data
    for var_name in ["value", "cause", "race", "county"]:
        assert df_obs[var_name].isna().sum() == 0, (
            "There are missing values in the "
            + var_name
            + " column of the observations."
        )
    assert len(df_obs[df_obs.duplicated(["cause", "race", "county"])]) == 0, (
        "There are duplicated rows in the observations."
    )
    count_obs = df_obs[["cause", "race", "county"]].value_counts()
    assert (len(count_obs.unique()) == 1) and (count_obs.unique()[0] == 1), (
        "There are missing combinations of cause, race and county in the observations."
    )

    # Check the margins data
    assert df_margins["cause"].isna().sum() == 0, (
        "There are missing values in the cause column of the margins."
    )
    assert df_margins["value_agg_over_race_county"].isna().sum() == 0, (
        "There are missing values in the value_agg_over_race_county column of the margins."
    )
    assert len(df_margins[df_margins.duplicated(["cause"])]) == 0, (
        "There are duplicated rows in the margins data frame."
    )

    # Check consistency between observations and margins
    assert len(df_obs["cause"].unique()) == len(df_margins["cause"].unique()), (
        "The number of categories for cause should be the same in the observations and margins data frames."
    )
    assert set(df_obs["cause"].unique().tolist()) == set(
        df_margins["cause"].unique().tolist()
    ), (
        "The names of the categories for cause should be the same in the observations and margins data frames."
    )

    # Create input variables for the raking functions
    df_obs.sort_values(by=["county", "race", "cause"], inplace=True)
    df_margins.sort_values(by=["cause"], inplace=True)
    I = len(df_obs["cause"].unique()) - 1
    J = len(df_obs["race"].unique()) - 1
    K = len(df_obs["county"].unique())
    y = df_obs.value.to_numpy()
    s = df_margins["value_agg_over_race_county"].to_numpy()
    if weights is not None:
        q = df_obs[weights].to_numpy()
    else:
        q = None
    if lower is not None:
        l = df_obs[lower].to_numpy()
    else:
        l = None
    if upper is not None:
        h = df_obs[upper].to_numpy()
    else:
        h = None
    return (y, s, I, J, K, q, l, h)



"""Module with methods to compute the covariance matrices of observations and margins"""

import numpy as np
import pandas as pd

pd.options.mode.chained_assignment = None


def check_observations(
    df_obs: pd.DataFrame, var_names: list, draws: str
) -> None:
    """Check whether the observations data frame is valid.

    Parameters
    ----------
    df_obs : pd.DataFrame
        Observations data
    var_names : list of strings
        Names of the variables over which we rake (e.g. cause, race, county)
    draws : string
        Names of the column containing the indices of the draws

    Returns
    -------
    None
    """
    assert isinstance(df_obs, pd.DataFrame), (
        "The observations should be a pandas data frame."
    )
    assert len(df_obs) >= 2, (
        "There should be at least 2 data points for the observations."
    )

    assert "value" in df_obs.columns.tolist(), (
        "The observations data frame should contain a value column."
    )

    assert isinstance(var_names, list), (
        "Please enter the names of the columns containing the values of the categorical variables as a list."
    )
    for var_name in var_names:
        assert isinstance(var_name, str), (
            "The name of the categorical variable "
            + str(var_name)
            + " should be a string."
        )
        assert var_name in df_obs.columns.tolist(), (
            "The column for the categorical variable "
            + var_name
            + " is missing from the observations data frame."
        )

    assert isinstance(draws, str), (
        "The name of the column containing the draws should be a string."
    )
    assert draws in df_obs.columns.tolist(), (
        "The column containing the draws is missing from the observations data frame."
    )

    assert df_obs.value.isna().sum() == 0, (
        "There are missing values in the value column of the observations."
    )
    for var_name in var_names:
        assert df_obs[var_name].isna().sum() == 0, (
            "There are missing values in the "
            + var_name
            + " column of the observations."
        )
    assert df_obs[draws].isna().sum() == 0, (
        "There are missing values in the draws column of the observations."
    )
    assert len(df_obs[df_obs.duplicated(var_names + [draws])]) == 0, (
        "There are duplicated rows in the observations."
    )
    count_obs = df_obs[var_names + [draws]].value_counts()
    assert (len(count_obs.unique()) == 1) and (count_obs.unique()[0] == 1), (
        "There are missing combinations of variables and draws in the observations."
    )


def compute_covariance_obs(
    df_obs: pd.DataFrame, var_names: list, draws: str
) -> np.ndarray:
    """Compute the covariance matrix of the observations.

    The observations will be sorted by var3, var2, var1, meaning that
    sigma_yy contains on its diagonal in this order the variances of
    y_111, ... , y_I11, y_121, ... , y_IJ1, y_112, ... , y_IJK.

    Parameters
    ----------
    df_obs : pd.DataFrame
        Observations data
    var_names : list of strings
        Names of the variables over which we rake (e.g. cause, race, county)
    draws : string
        Names of the column containing the indices of the draws

    Returns
    -------
    sigma_yy : np.ndarray
        (I * J * K) * (I * J * K) covariance matrix
    """
    check_observations(df_obs, var_names, draws)

    nsamples = len(df_obs[draws].unique())
    var_names_reverse = var_names.copy()
    var_names_reverse.reverse()
    df = df_obs[["value"] + var_names + [draws]]
    df.sort_values(by=var_names_reverse + [draws], inplace=True)
    value = df["value"].to_numpy()
    X = np.reshape(value, shape=(nsamples, -1), order="F")
    Xmean = np.mean(X, axis=0)
    Xc = X - Xmean
    sigma_yy = np.matmul(np.transpose(Xc), Xc) / nsamples
    return sigma_yy


def check_margins_1D(
    df_margins: pd.DataFrame, var_names: list, draws: str
) -> None:
    """Check whether the margin data frame is valid in 1D.

    Parameters
    ----------
    df_margins : pd.DataFrame
        Margins data (sums over the first variable)
    var_names : list of strings
        Names of the variables over which we rake (e.g. cause, race, county)
    draws : string
        Names of the column containing the indices of the draws

    Returns
    -------
    None
    """
    assert isinstance(df_margins, pd.DataFrame), (
        "The margins should be a pandas data frame."
    )
    assert len(df_margins) >= 1, (
        "There should be at least 1 data point for the margins."
    )

    assert isinstance(var_names, list), (
        "Please enter the names of the columns containing the values of the categorical variables as a list."
    )
    assert len(var_names) == 1, "You should have 1 categorical variable."
    assert isinstance(var_names[0], str), (
        "The name of the categorical variable should be a string."
    )
    assert "value_agg_over_" + var_names[0] in df_margins.columns.tolist(), (
        "The column for the aggregated value over "
        + var_names[0]
        + " is missing from the margins data frame."
    )

    assert isinstance(draws, str), (
        "The name of the column containing the draws should be a string."
    )
    assert draws in df_margins.columns.tolist(), (
        "The column containing the draws is missing from the margins data frame."
    )

    assert df_margins["value_agg_over_" + var_names[0]].isna().sum() == 0, (
        "There are missing values in the value_agg_over"
        + var_names[0]
        + " column of the margins."
    )
    assert df_margins[draws].isna().sum() == 0, (
        "There are missing values in the draws column of the margins."
    )
    assert len(df_margins[df_margins.duplicated([draws])]) == 0, (
        "There are duplicated rows in the margins."
    )
    count_obs = df_margins[[draws]].value_counts()
    assert (len(count_obs.unique()) == 1) and (count_obs.unique()[0] == 1), (
        "There are missing draws in the margins."
    )


def compute_covariance_margins_1D(
    df_margins: pd.DataFrame, var_names: list, draws: str
) -> np.ndarray:
    """Compute the covariance matrix of the margins in 1D.

    Parameters
    ----------
    df_margins : pd.DataFrame
        Margins data (sums over the first variable)
    var_names : list of strings
        Names of the variables over which we rake (e.g. cause, race, county)
    draws : string
        Names of the column containing the indices of the draws

    Returns
    -------
    sigma_ss : np.ndarray
        1 * 1 covariance matrix
    """
    check_margins_1D(df_margins, var_names, draws)

    nsamples = len(df_margins[draws].unique())
    df = df_margins[["value_agg_over_" + var_names[0]] + [draws]]
    df.sort_values(by=[draws], inplace=True)
    value = df["value_agg_over_" + var_names[0]].to_numpy()
    X = np.reshape(value, shape=(nsamples, -1), order="F")
    Xmean = np.mean(X, axis=0)
    Xc = X - Xmean
    sigma_ss = np.matmul(np.transpose(Xc), Xc) / nsamples
    return sigma_ss


def check_margins_2D(
    df_margins_1: pd.DataFrame,
    df_margins_2: pd.DataFrame,
    var_names: list,
    draws: str,
) -> None:
    """Check whether the margins data frame are valid in 2D.

    Parameters
    ----------
    df_margins_1 : pd.DataFrame
        Margins data (sums over the first variable)
    df_margins_2 : pd.DataFrame
        Margins data (sums over the second variable)
    var_names : list of strings
        Names of the variables over which we rake (e.g. cause, race, county)
    draws : string
        Names of the column containing the indices of the draws

    Returns
    -------
    None
    """
    assert isinstance(df_margins_1, pd.DataFrame), (
        "The margins for the first variable should be a pandas data frame."
    )
    assert len(df_margins_1) >= 2, (
        "There should be at least 2 data points for the first margins."
    )

    assert isinstance(df_margins_2, pd.DataFrame), (
        "The margins for the second variable should be a pandas data frame."
    )
    assert len(df_margins_2) >= 2, (
        "There should be at least 2 data points for the second margins."
    )

    assert isinstance(var_names, list), (
        "Please enter the names of the columns containing the values of the categorical variables as a list."
    )
    assert len(var_names) == 2, "You should have 2 categorical variables."
    for var_name in var_names:
        assert isinstance(var_name, str), (
            "The name of the categorical variable "
            + str(var_name)
            + " should be a string."
        )

    assert var_names[1] in df_margins_1.columns.tolist(), (
        "The column for the categorigal variable "
        + var_name[1]
        + " is missing from the first margins data frame."
    )
    assert "value_agg_over_" + var_names[0] in df_margins_1.columns.tolist(), (
        "The column for the aggregated value over "
        + var_names[0]
        + " is missing from the first margins data frame."
    )

    assert var_names[0] in df_margins_2.columns.tolist(), (
        "The column for the categorigal variable "
        + var_name[0]
        + " is missing from the second margins data frame."
    )
    assert "value_agg_over_" + var_names[1] in df_margins_2.columns.tolist(), (
        "The column for the aggregated value over "
        + var_names[1]
        + " is missing from the second margins data frame."
    )

    assert isinstance(draws, str), (
        "The name of the column containing the draws should be a string."
    )
    assert draws in df_margins_1.columns.tolist(), (
        "The column containing the draws is missing from the first margins data frame."
    )
    assert draws in df_margins_2.columns.tolist(), (
        "The column containing the draws is missing from the second margins data frame."
    )

    assert df_margins_1[var_names[1]].isna().sum() == 0, (
        "There are missing values in the "
        + var_names[1]
        + " column of the margins."
    )
    assert df_margins_1["value_agg_over_" + var_names[0]].isna().sum() == 0, (
        "There are missing values in the value_agg_over"
        + var_names[0]
        + " column of the margins."
    )
    assert df_margins_1[draws].isna().sum() == 0, (
        "There are missing values in the draws column of the first margins."
    )
    assert (
        len(df_margins_1[df_margins_1.duplicated([var_names[1], draws])]) == 0
    ), "There are duplicated rows in the first margins data frame."
    count_obs = df_margins_1[[var_names[1], draws]].value_counts()
    assert (len(count_obs.unique()) == 1) and (count_obs.unique()[0] == 1), (
        "There are missing combinations of "
        + var_names[1]
        + " and draws in the first margins."
    )

    assert df_margins_2[var_names[0]].isna().sum() == 0, (
        "There are missing values in the "
        + var_names[0]
        + " column of the margins."
    )
    assert df_margins_2["value_agg_over_" + var_names[1]].isna().sum() == 0, (
        "There are missing values in the value_agg_over"
        + var_names[1]
        + " column of the margins."
    )
    assert df_margins_2[draws].isna().sum() == 0, (
        "There are missing values in the draws column of the second margins."
    )
    assert (
        len(df_margins_2[df_margins_2.duplicated([var_names[0], draws])]) == 0
    ), "There are duplicated rows in the second margins data frame."
    count_obs = df_margins_2[[var_names[0], draws]].value_counts()
    assert (len(count_obs.unique()) == 1) and (count_obs.unique()[0] == 1), (
        "There are missing combinations of "
        + var_names[0]
        + " and draws in the second margins."
    )


def compute_covariance_margins_2D(
    df_margins_1: pd.DataFrame,
    df_margins_2: pd.DataFrame,
    var_names: list,
    draws: str,
) -> np.ndarray:
    """Compute the covariance matrix of the margins in 2D.

    The margins are sorted in the same order as what is done
    when computing the constraint matrix.

    Parameters
    ----------
    df_margins_1 : pd.DataFrame
        Margins data (sums over the first variable)
    df_margins_2 : pd.DataFrame
        Margins data (sums over the second variable)
    var_names : list of strings
        Names of the variables over which we rake (e.g. cause, race, county)
    draws : string
        Names of the column containing the indices of the draws

    Returns
    -------
    sigma_ss : np.ndarray
        (I + J - 1) * (I + J - 1) covariance matrix
    """
    check_margins_2D(df_margins_1, df_margins_2, var_names, draws)

    nsamples = len(df_margins_1[draws].unique())
    df1 = df_margins_1[[var_names[1], "value_agg_over_" + var_names[0], draws]]
    df1.sort_values(by=[var_names[1], draws], inplace=True)
    df2 = df_margins_2[[var_names[0], "value_agg_over_" + var_names[1], draws]]
    df2.sort_values(by=[var_names[0], draws], inplace=True)
    value1 = df1["value_agg_over_" + var_names[0]].to_numpy()
    value2 = df2["value_agg_over_" + var_names[1]].to_numpy()
    value = np.concatenate((value1, value2))
    X = np.reshape(value, shape=(nsamples, -1), order="F")
    X = X[:, 0:-1]
    Xmean = np.mean(X, axis=0)
    Xc = X - Xmean
    sigma_ss = np.matmul(np.transpose(Xc), Xc) / nsamples
    return sigma_ss


def check_margins_3D(
    df_margins_1: pd.DataFrame,
    df_margins_2: pd.DataFrame,
    df_margins_3: pd.DataFrame,
    var_names: list,
    draws: str,
) -> None:
    """Check whether the margins data frames are valid in 3D.

    Parameters
    ----------
    df_margins_1 : pd.DataFrame
        Margins data (sums over the first variable)
    df_margins_2 : pd.DataFrame
        Margins data (sums over the second variable)
    df_margins_3 : pd.DataFrame
        Margins data (sums over the third variable)
    var_names : list of strings
        Names of the variables over which we rake (e.g. cause, race, county)
    draws : string
        Names of the column containing the indices of the draws

    Returns
    -------
    None
    """
    assert isinstance(df_margins_1, pd.DataFrame), (
        "The margins for the first variable should be a pandas data frame."
    )
    assert len(df_margins_1) >= 4, (
        "There should be at least 4 data points for the first margins."
    )

    assert isinstance(df_margins_2, pd.DataFrame), (
        "The margins for the second variable should be a pandas data frame."
    )
    assert len(df_margins_2) >= 4, (
        "There should be at least 4 data points for the second margins."
    )

    assert isinstance(df_margins_3, pd.DataFrame), (
        "The margins for the second variable should be a pandas data frame."
    )
    assert len(df_margins_3) >= 4, (
        "There should be at least 4 data points for the third margins."
    )

    assert isinstance(var_names, list), (
        "Please enter the names of the columns containing the values of the categorical variables as a list."
    )
    assert len(var_names) == 3, "You should have 3 categorical variables."
    for var_name in var_names:
        assert isinstance(var_name, str), (
            "The name of the categorical variable "
            + str(var_name)
            + " should be a string."
        )

    assert var_names[1] in df_margins_1.columns.tolist(), (
        "The column for the categorigal variable "
        + var_name[1]
        + " is missing from the first margins data frame."
    )
    assert var_names[2] in df_margins_1.columns.tolist(), (
        "The column for the categorigal variable "
        + var_name[2]
        + " is missing from the first margins data frame."
    )
    assert "value_agg_over_" + var_names[0] in df_margins_1.columns.tolist(), (
        "The column for the aggregated value over "
        + var_names[0]
        + " is missing from the first margins data frame."
    )

    assert var_names[0] in df_margins_2.columns.tolist(), (
        "The column for the categorigal variable "
        + var_name[0]
        + " is missing from the second margins data frame."
    )
    assert var_names[2] in df_margins_2.columns.tolist(), (
        "The column for the categorigal variable "
        + var_name[2]
        + " is missing from the second margins data frame."
    )
    assert "value_agg_over_" + var_names[1] in df_margins_2.columns.tolist(), (
        "The column for the aggregated value over "
        + var_names[1]
        + " is missing from the second margins data frame."
    )

    assert var_names[0] in df_margins_3.columns.tolist(), (
        "The column for the categorigal variable "
        + var_name[0]
        + " is missing from the third margins data frame."
    )
    assert var_names[1] in df_margins_3.columns.tolist(), (
        "The column for the categorigal variable "
        + var_name[1]
        + " is missing from the third margins data frame."
    )
    assert "value_agg_over_" + var_names[2] in df_margins_3.columns.tolist(), (
        "The column for the aggregated value over "
        + var_names[2]
        + " is missing from the third margins data frame."
    )

    assert isinstance(draws, str), (
        "The name of the column containing the draws should be a string."
    )
    assert draws in df_margins_1.columns.tolist(), (
        "The column containing the draws is missing from the first margins data frame."
    )
    assert draws in df_margins_2.columns.tolist(), (
        "The column containing the draws is missing from the second margins data frame."
    )
    assert draws in df_margins_3.columns.tolist(), (
        "The column containing the draws is missing from the third margins data frame."
    )

    assert df_margins_1[var_names[1]].isna().sum() == 0, (
        "There are missing values in the "
        + var_names[1]
        + " column of the margins."
    )
    assert df_margins_1[var_names[2]].isna().sum() == 0, (
        "There are missing values in the "
        + var_names[2]
        + " column of the margins."
    )
    assert df_margins_1["value_agg_over_" + var_names[0]].isna().sum() == 0, (
        "There are missing values in the value_agg_over"
        + var_names[0]
        + " column of the margins."
    )
    assert df_margins_1[draws].isna().sum() == 0, (
        "There are missing values in the draws column of the first margins."
    )
    assert (
        len(
            df_margins_1[
                df_margins_1.duplicated([var_names[1], var_names[2], draws])
            ]
        )
        == 0
    ), "There are duplicated rows in the first margins data frame."
    count_obs = df_margins_1[[var_names[1], var_names[2], draws]].value_counts()
    assert (len(count_obs.unique()) == 1) and (count_obs.unique()[0] == 1), (
        "There are missing combinations of "
        + var_names[1]
        + ", "
        + var_names[2]
        + " and draws in the first margins."
    )

    assert df_margins_2[var_names[0]].isna().sum() == 0, (
        "There are missing values in the "
        + var_names[0]
        + " column of the margins."
    )
    assert df_margins_2[var_names[2]].isna().sum() == 0, (
        "There are missing values in the "
        + var_names[2]
        + " column of the margins."
    )
    assert df_margins_2["value_agg_over_" + var_names[1]].isna().sum() == 0, (
        "There are missing values in the value_agg_over"
        + var_names[1]
        + " column of the margins."
    )
    assert df_margins_2[draws].isna().sum() == 0, (
        "There are missing values in the draws column of the second margins."
    )
    assert (
        len(
            df_margins_2[
                df_margins_2.duplicated([var_names[0], var_names[2], draws])
            ]
        )
        == 0
    ), "There are duplicated rows in the second margins data frame."
    count_obs = df_margins_2[[var_names[0], var_names[2], draws]].value_counts()
    assert (len(count_obs.unique()) == 1) and (count_obs.unique()[0] == 1), (
        "There are missing combinations of "
        + var_names[0]
        + ", "
        + var_names[2]
        + " and draws in the second margins."
    )

    assert df_margins_3[var_names[0]].isna().sum() == 0, (
        "There are missing values in the "
        + var_names[0]
        + " column of the margins."
    )
    assert df_margins_3[var_names[1]].isna().sum() == 0, (
        "There are missing values in the "
        + var_names[1]
        + " column of the margins."
    )
    assert df_margins_3["value_agg_over_" + var_names[2]].isna().sum() == 0, (
        "There are missing values in the value_agg_over"
        + var_names[2]
        + " column of the margins."
    )
    assert df_margins_3[draws].isna().sum() == 0, (
        "There are missing values in the draws column of the third margins."
    )
    assert (
        len(
            df_margins_3[
                df_margins_3.duplicated([var_names[0], var_names[1], draws])
            ]
        )
        == 0
    ), "There are duplicated rows in the third margins data frame."
    count_obs = df_margins_3[[var_names[0], var_names[1], draws]].value_counts()
    assert (len(count_obs.unique()) == 1) and (count_obs.unique()[0] == 1), (
        "There are missing combinations of "
        + var_names[0]
        + ", "
        + var_names[1]
        + " and draws in the third margins."
    )


def compute_covariance_margins_3D(
    df_margins_1: pd.DataFrame,
    df_margins_2: pd.DataFrame,
    df_margins_3: pd.DataFrame,
    var_names: list,
    draws: str,
) -> np.ndarray:
    """Compute the covariance matrix of the margins in 3D.

    The margins are sorted in the same order as what is done
    when computing the constraint matrix.

    Parameters
    ----------
    df_margins_1 : pd.DataFrame
        Margins data (sums over the first variable)
    df_margins_2 : pd.DataFrame
        Margins data (sums over the second variable)
    df_margins_3 : pd.DataFrame
        Margins data (sums over the third variable)
    var_names : list of strings
        Names of the variables over which we rake (e.g. cause, race, county)
    draws : string
        Names of the column containing the indices of the draws

    Returns
    -------
    sigma_ss : np.ndarray
        (I J + I K + J K - I - J - K + 1) * (I J + I K + J K - I - J - K + 1) covariance matrix
    """
    check_margins_3D(df_margins_1, df_margins_2, df_margins_3, var_names, draws)

    nsamples = len(df_margins_1[draws].unique())
    var1 = df_margins_2[var_names[0]].unique().tolist()
    var2 = df_margins_1[var_names[1]].unique().tolist()
    var3 = df_margins_1[var_names[2]].unique().tolist()
    var1.sort()
    var2.sort()
    var3.sort()
    df1 = df_margins_1[
        [var_names[1], var_names[2], "value_agg_over_" + var_names[0], draws]
    ]
    df1 = df1.loc[
        (df1[var_names[1]].isin(var2[0:-1]))
        | ((df1[var_names[1]] == var2[-1]) & (df1[var_names[2]] == var3[-1]))
    ]
    df1.sort_values(by=[var_names[2], var_names[1], draws], inplace=True)
    df2 = df_margins_2[
        [var_names[0], var_names[2], "value_agg_over_" + var_names[1], draws]
    ]
    df2 = df2.loc[df2[var_names[2]].isin(var3[0:-1])]
    df2.sort_values(by=[var_names[0], var_names[2], draws], inplace=True)
    df3 = df_margins_3[
        [var_names[0], var_names[1], "value_agg_over_" + var_names[2], draws]
    ]
    df3 = df3.loc[df3[var_names[0]].isin(var1[0:-1])]
    df3.sort_values(by=[var_names[1], var_names[0], draws], inplace=True)
    value1 = df1["value_agg_over_" + var_names[0]].to_numpy()
    value2 = df2["value_agg_over_" + var_names[1]].to_numpy()
    value3 = df3["value_agg_over_" + var_names[2]].to_numpy()
    value = np.concatenate((value1, value2, value3))
    X = np.reshape(value, shape=(nsamples, -1), order="F")
    Xmean = np.mean(X, axis=0)
    Xc = X - Xmean
    sigma_ss = np.matmul(np.transpose(Xc), Xc) / nsamples
    return sigma_ss


def check_margins_USHD(
    df_margins: pd.DataFrame, var_names: list, draws: str
) -> None:
    """Check whether the margin data frame is valid for the USHD case.

    Parameters
    ----------
    df_margins : pd.DataFrame
        Margins data (sums over the first variable)
    var_names : list of strings
        Names of the variables over which we rake (e.g. cause)
    draws : string
        Names of the column containing the indices of the draws

    Returns
    -------
    None
    """
    assert isinstance(df_margins, pd.DataFrame), (
        "The margins should be a pandas data frame."
    )
    assert len(df_margins) >= 2, (
        "There should be at least 2 data points for the margins."
    )

    assert "value_agg_over_race_county" in df_margins.columns.tolist(), (
        "The column for the aggregated value over race and county is missing from the margins data frame."
    )
    assert df_margins["value_agg_over_race_county"].isna().sum() == 0, (
        "There are missing values in the value_agg_over_race_county column of the margins."
    )

    assert isinstance(draws, str), (
        "The name of the column containing the draws should be a string."
    )
    assert draws in df_margins.columns.tolist(), (
        "The column containing the draws is missing from the margins data frame."
    )
    assert df_margins[draws].isna().sum() == 0, (
        "There are missing values in the draws column of the margins."
    )
    assert len(df_margins[df_margins.duplicated(["cause", draws])]) == 0, (
        "There are duplicated rows in the margins."
    )
    count_obs = df_margins[["cause", draws]].value_counts()
    assert (len(count_obs.unique()) == 1) and (count_obs.unique()[0] == 1), (
        "There are missing draws in the margins."
    )


def compute_covariance_margins_USHD(
    df_margins: pd.DataFrame,
    var_names: list,
    draws: str,
    I: int,
    J: int,
    K: int,
) -> np.ndarray:
    """Compute the covariance matrix of the margins for the USHD case.

    Parameters
    ----------
    df_margins : pd.DataFrame
        Margins data (sums over the first variable)
    var_names : list of strings
        Names of the variables over which we rake (e.g. cause)
    draws : string
        Names of the column containing the indices of the draws
    I : int
        Number of causes of deaths
    J : int
        Number of races and ethnicities
    K : int
        Number of counties

    Returns
    -------
    sigma_ss : np.ndarray
        I * I covariance matrix
    """
    check_margins_USHD(df_margins, var_names, draws)

    nsamples = len(df_margins[draws].unique())
    df = df_margins[["cause", "value_agg_over_race_county", draws]].loc[
        df_margins.cause != "_all"
    ]
    df.sort_values(by=["cause", draws], inplace=True)
    value = df["value_agg_over_race_county"].to_numpy()
    X = np.reshape(value, shape=(nsamples, -1), order="F")
    Xmean = np.mean(X, axis=0)
    Xc = X - Xmean
    sigma_ss = np.matmul(np.transpose(Xc), Xc) / nsamples
    sigma_12 = np.zeros((I, 2 * K + J * K + (I - 1) * K))
    sigma_22 = np.zeros(
        (2 * K + J * K + (I - 1) * K, 2 * K + J * K + (I - 1) * K)
    )
    sigma_ss = np.concatenate(
        (
            np.concatenate((sigma_ss, sigma_12), axis=1),
            np.concatenate((np.transpose(sigma_12), sigma_22), axis=1),
        ),
        axis=0,
    )
    return sigma_ss


def check_obs_margins_1D(
    df_obs: pd.DataFrame, df_margins: pd.DataFrame, draws: str
) -> None:
    """Check whether the observations and margins data frames are consistent in 1D.

    Parameters
    ----------
    df_obs : pd.DataFrame
        Observations data
    df_margins : pd.DataFrame
        Margins data (sums over the first variable)
    draws : string
        Names of the column containing the indices of the draws

    Returns
    -------
    None
    """
    assert set(df_obs[draws].unique().tolist()) == set(
        df_margins[draws].unique().tolist()
    ), (
        "The draws should be the same in the observations and margins data frames."
    )


def compute_covariance_obs_margins_1D(
    df_obs: pd.DataFrame, df_margins: pd.DataFrame, var_names: list, draws: str
) -> np.ndarray:
    """Compute the covariance matrix of the observations and the margins in 1D.

    The observations will be sorted by var3, var2, var1, meaning that
    sigma_yy contains on its diagonal in this order the variances of
    y_111, ... , y_I11, y_121, ... , y_IJ1, y_112, ... , y_IJK.

    Parameters
    ----------
    df_obs : pd.DataFrame
        Observations data
    df_margins : pd.DataFrame
        Margins data (sums over the first variable)
    var_names : list of strings
        Names of the variables over which we rake (e.g. cause, race, county)
    draws : string
        Names of the column containing the indices of the draws

    Returns
    -------
    sigma_ys : np.ndarray
        (I * J * K) * 1 covariance matrix
    """
    check_observations(df_obs, var_names, draws)
    check_margins_1D(df_margins, var_names, draws)
    check_obs_margins_1D(df_obs, df_margins, draws)

    nsamples = len(df_obs[draws].unique())
    var_names_reverse = var_names.copy()
    var_names_reverse.reverse()
    df_obs = df_obs[["value"] + var_names + [draws]]
    df_obs.sort_values(by=var_names_reverse + [draws], inplace=True)
    df_margins = df_margins[["value_agg_over_" + var_names[0]] + [draws]]
    df_margins.sort_values(by=[draws], inplace=True)
    value_obs = df_obs["value"].to_numpy()
    X = np.reshape(value_obs, shape=(nsamples, -1), order="F")
    value_margins = df_margins["value_agg_over_" + var_names[0]].to_numpy()
    Y = np.reshape(value_margins, shape=(nsamples, -1), order="F")
    Xmean = np.mean(X, axis=0)
    Ymean = np.mean(Y, axis=0)
    Xc = X - Xmean
    Yc = Y - Ymean
    sigma_ys = np.matmul(np.transpose(Xc), Yc) / nsamples
    return sigma_ys


def check_obs_margins_2D(
    df_obs: pd.DataFrame,
    df_margins_1: pd.DataFrame,
    df_margins_2: pd.DataFrame,
    var_names: list,
    draws: str,
) -> None:
    """Check whether the observations and margins data frames are consistent in 2D.

    Parameters
    ----------
    df_obs : pd.DataFrame
        Observations data
    df_margins_1 : pd.DataFrame
        Margins data (sums over the first variable)
    df_margins_2 : pd.DataFrame
        Margins data (sums over the second variable)
    var_names : list of strings
        Names of the variables over which we rake (e.g. cause, race, county)
    draws : string
        Names of the column containing the indices of the draws

    Returns
    -------
    None
    """
    assert set(df_obs[draws].unique().tolist()) == set(
        df_margins_1[draws].unique().tolist()
    ), (
        "The draws should be the same in the observations and the first margins data frames."
    )

    assert set(df_obs[draws].unique().tolist()) == set(
        df_margins_2[draws].unique().tolist()
    ), (
        "The draws should be the same in the observations and the second margins data frames."
    )

    assert len(df_obs[var_names[0]].unique()) == len(
        df_margins_2[var_names[0]].unique()
    ), (
        "The number of categories for "
        + var_names[0]
        + " should be the same in the observations and margins data frames."
    )
    assert set(df_obs[var_names[0]].unique().tolist()) == set(
        df_margins_2[var_names[0]].unique().tolist()
    ), (
        "The names of the categories for "
        + var_names[0]
        + " should be the same in the observations and margins data frames."
    )
    assert len(df_obs[var_names[1]].unique()) == len(
        df_margins_1[var_names[1]].unique()
    ), (
        "The number of categories for "
        + var_names[1]
        + " should be the same in the observations and margins data frames."
    )
    assert set(df_obs[var_names[1]].unique().tolist()) == set(
        df_margins_1[var_names[1]].unique().tolist()
    ), (
        "The names of the categories for "
        + var_names[1]
        + " should be the same in the observations and margins data frames."
    )


def compute_covariance_obs_margins_2D(
    df_obs: pd.DataFrame,
    df_margins_1: pd.DataFrame,
    df_margins_2: pd.DataFrame,
    var_names: list,
    draws: str,
) -> np.ndarray:
    """Compute the covariance matrix of the observations and the margins in 2D.

    The observations will be sorted by var3, var2, var1, meaning that
    sigma_yy contains on its diagonal in this order the variances of
    y_111, ... , y_I11, y_121, ... , y_IJ1, y_112, ... , y_IJK.

    Parameters
    ----------
    df_obs : pd.DataFrame
        Observations data
    df_margins_1 : pd.DataFrame
        Margins data (sums over the first variable)
    df_margins_2 : pd.DataFrame
        Margins data (sums over the second variable)
    var_names : list of strings
        Names of the variables over which we rake (e.g. cause, race, county)
    draws : string
        Names of the column containing the indices of the draws

    Returns
    -------
    sigma_ys : np.ndarray
        (I * J * K) * (I + J - 1) covariance matrix
    """
    check_observations(df_obs, var_names, draws)
    check_margins_2D(df_margins_1, df_margins_2, var_names, draws)
    check_obs_margins_2D(df_obs, df_margins_1, df_margins_2, var_names, draws)

    nsamples = len(df_obs[draws].unique())
    var_names_reverse = var_names.copy()
    var_names_reverse.reverse()
    df_obs = df_obs[["value"] + var_names + [draws]]
    df_obs.sort_values(by=var_names_reverse + [draws], inplace=True)
    df_margins_1 = df_margins_1[
        [var_names[1], "value_agg_over_" + var_names[0], draws]
    ]
    df_margins_1.sort_values(by=[var_names[1], draws], inplace=True)
    df_margins_2 = df_margins_2[
        [var_names[0], "value_agg_over_" + var_names[1], draws]
    ]
    df_margins_2.sort_values(by=[var_names[0], draws], inplace=True)
    value_obs = df_obs["value"].to_numpy()
    X = np.reshape(value_obs, shape=(nsamples, -1), order="F")
    value_margins_1 = df_margins_1["value_agg_over_" + var_names[0]].to_numpy()
    value_margins_2 = df_margins_2["value_agg_over_" + var_names[1]].to_numpy()
    value_margins = np.concatenate((value_margins_1, value_margins_2))
    Y = np.reshape(value_margins, shape=(nsamples, -1), order="F")
    Y = Y[:, 0:-1]
    Xmean = np.mean(X, axis=0)
    Ymean = np.mean(Y, axis=0)
    Xc = X - Xmean
    Yc = Y - Ymean
    sigma_ys = np.matmul(np.transpose(Xc), Yc) / nsamples
    return sigma_ys


def check_obs_margins_3D(
    df_obs: pd.DataFrame,
    df_margins_1: pd.DataFrame,
    df_margins_2: pd.DataFrame,
    df_margins_3: pd.DataFrame,
    var_names: list,
    draws: str,
) -> None:
    """Check whether the observations and margins data frame are consistent in 3D.

    Parameters
    ----------
    df_obs : pd.DataFrame
        Observations data
    df_margins_1 : pd.DataFrame
        Margins data (sums over the first variable)
    df_margins_2 : pd.DataFrame
        Margins data (sums over the second variable)
    df_margins_3 : pd.DataFrame
        Margins data (sums over the third variable)
    var_names : list of strings
        Names of the variables over which we rake (e.g. cause, race, county)
    draws : string
        Names of the column containing the indices of the draws

    Returns
    -------
    None
    """
    assert set(df_obs[draws].unique().tolist()) == set(
        df_margins_1[draws].unique().tolist()
    ), (
        "The draws should be the same in the observations and the first margins data frames."
    )

    assert set(df_obs[draws].unique().tolist()) == set(
        df_margins_2[draws].unique().tolist()
    ), (
        "The draws should be the same in the observations and the second margins data frames."
    )

    assert set(df_obs[draws].unique().tolist()) == set(
        df_margins_3[draws].unique().tolist()
    ), (
        "The draws should be the same in the observations and the third margins data frames."
    )

    assert set(df_obs[draws].unique().tolist()) == set(
        df_margins_2[draws].unique().tolist()
    ), (
        "The draws should be the same in the observations and the second margins data frames."
    )

    assert len(df_obs[var_names[1]].unique()) == len(
        df_margins_1[var_names[1]].unique()
    ), (
        "The number of categories for "
        + var_names[1]
        + " should be the same in the observations and first margins data frames."
    )
    assert set(df_obs[var_names[1]].unique().tolist()) == set(
        df_margins_1[var_names[1]].unique().tolist()
    ), (
        "The names of the categories for "
        + var_names[1]
        + " should be the same in the observations and first margins data frames."
    )
    assert len(df_obs[var_names[2]].unique()) == len(
        df_margins_1[var_names[2]].unique()
    ), (
        "The number of categories for "
        + var_names[2]
        + " should be the same in the observations and first margins data frames."
    )
    assert set(df_obs[var_names[2]].unique().tolist()) == set(
        df_margins_1[var_names[2]].unique().tolist()
    ), (
        "The names of the categories for "
        + var_names[2]
        + " should be the same in the observations and first margins data frames."
    )

    assert len(df_obs[var_names[0]].unique()) == len(
        df_margins_2[var_names[0]].unique()
    ), (
        "The number of categories for "
        + var_names[0]
        + " should be the same in the observations and second margins data frames."
    )
    assert set(df_obs[var_names[0]].unique().tolist()) == set(
        df_margins_2[var_names[0]].unique().tolist()
    ), (
        "The names of the categories for "
        + var_names[0]
        + " should be the same in the observations and second margins data frames."
    )
    assert len(df_obs[var_names[2]].unique()) == len(
        df_margins_2[var_names[2]].unique()
    ), (
        "The number of categories for "
        + var_names[2]
        + " should be the same in the observations and second margins data frames."
    )
    assert set(df_obs[var_names[2]].unique().tolist()) == set(
        df_margins_2[var_names[2]].unique().tolist()
    ), (
        "The names of the categories for "
        + var_names[2]
        + " should be the same in the observations and second margins data frames."
    )

    assert len(df_obs[var_names[0]].unique()) == len(
        df_margins_3[var_names[0]].unique()
    ), (
        "The number of categories for "
        + var_names[0]
        + " should be the same in the observations and third margins data frames."
    )
    assert set(df_obs[var_names[0]].unique().tolist()) == set(
        df_margins_3[var_names[0]].unique().tolist()
    ), (
        "The names of the categories for "
        + var_names[0]
        + " should be the same in the observations and third margins data frames."
    )
    assert len(df_obs[var_names[1]].unique()) == len(
        df_margins_3[var_names[1]].unique()
    ), (
        "The number of categories for "
        + var_names[1]
        + " should be the same in the observations and third margins data frames."
    )
    assert set(df_obs[var_names[1]].unique().tolist()) == set(
        df_margins_3[var_names[1]].unique().tolist()
    ), (
        "The names of the categories for "
        + var_names[1]
        + " should be the same in the observations and third margins data frames."
    )


def compute_covariance_obs_margins_3D(
    df_obs: pd.DataFrame,
    df_margins_1: pd.DataFrame,
    df_margins_2: pd.DataFrame,
    df_margins_3: pd.DataFrame,
    var_names: list,
    draws: str,
) -> np.ndarray:
    """Compute the covariance matrix of the observations and the margins in 3D.

    The observations will be sorted by var3, var2, var1, meaning that
    sigma_yy contains on its diagonal in this order the variances of
    y_111, ... , y_I11, y_121, ... , y_IJ1, y_112, ... , y_IJK.

    Parameters
    ----------
    df_obs : pd.DataFrame
        Observations data
    df_margins_1 : pd.DataFrame
        Margins data (sums over the first variable)
    df_margins_2 : pd.DataFrame
        Margins data (sums over the second variable)
    df_margins_3 : pd.DataFrame
        Margins data (sums over the third variable)
    var_names : list of strings
        Names of the variables over which we rake (e.g. cause, race, county)
    draws : string
        Names of the column containing the indices of the draws

    Returns
    -------
    sigma_ys : np.ndarray
        (I * J * K) * (I J + I K + J K - I - J - K + 1) covariance matrix
    """
    check_observations(df_obs, var_names, draws)
    check_margins_3D(df_margins_1, df_margins_2, df_margins_3, var_names, draws)
    check_obs_margins_3D(
        df_obs, df_margins_1, df_margins_2, df_margins_3, var_names, draws
    )

    nsamples = len(df_obs[draws].unique())
    var_names_reverse = var_names.copy()
    var_names_reverse.reverse()
    df_obs = df_obs[["value"] + var_names + [draws]]
    df_obs.sort_values(by=var_names_reverse + [draws], inplace=True)
    var1 = df_margins_2[var_names[0]].unique().tolist()
    var2 = df_margins_1[var_names[1]].unique().tolist()
    var3 = df_margins_1[var_names[2]].unique().tolist()
    var1.sort()
    var2.sort()
    var3.sort()
    df_margins_1 = df_margins_1[
        [var_names[1], var_names[2], "value_agg_over_" + var_names[0], draws]
    ]
    df_margins_1 = df_margins_1.loc[
        (df_margins_1[var_names[1]].isin(var2[0:-1]))
        | (
            (df_margins_1[var_names[1]] == var2[-1])
            & (df_margins_1[var_names[2]] == var3[-1])
        )
    ]
    df_margins_1.sort_values(
        by=[var_names[2], var_names[1], draws], inplace=True
    )
    df_margins_2 = df_margins_2[
        [var_names[0], var_names[2], "value_agg_over_" + var_names[1], draws]
    ]
    df_margins_2 = df_margins_2.loc[df_margins_2[var_names[2]].isin(var3[0:-1])]
    df_margins_2.sort_values(
        by=[var_names[0], var_names[2], draws], inplace=True
    )
    df_margins_3 = df_margins_3[
        [var_names[0], var_names[1], "value_agg_over_" + var_names[2], draws]
    ]
    df_margins_3 = df_margins_3.loc[df_margins_3[var_names[0]].isin(var1[0:-1])]
    df_margins_3.sort_values(
        by=[var_names[1], var_names[0], draws], inplace=True
    )
    value_obs = df_obs["value"].to_numpy()
    X = np.reshape(value_obs, shape=(nsamples, -1), order="F")
    value_margins_1 = df_margins_1["value_agg_over_" + var_names[0]].to_numpy()
    value_margins_2 = df_margins_2["value_agg_over_" + var_names[1]].to_numpy()
    value_margins_3 = df_margins_3["value_agg_over_" + var_names[2]].to_numpy()
    value_margins = np.concatenate(
        (value_margins_1, value_margins_2, value_margins_3)
    )
    Y = np.reshape(value_margins, shape=(nsamples, -1), order="F")
    Xmean = np.mean(X, axis=0)
    Ymean = np.mean(Y, axis=0)
    Xc = X - Xmean
    Yc = Y - Ymean
    sigma_ys = np.matmul(np.transpose(Xc), Yc) / nsamples
    return sigma_ys


def check_format_covariance(
    sigma_yy: np.ndarray, sigma_ss: np.ndarray, sigma_ys: np.ndarray
) -> None:
    """Check the format of the input covariance matrices.

    Parameters
    ----------
    sigma_yy : np.ndarray
        Covariance matrix of the observations
    sigma_ss : np.ndarray
        Covariance matrix of the margins
    sigma_ys : np.ndarray
        Covariance matrix of the observations and margins

    Returns
    -------
    None
    """
    assert isinstance(sigma_yy, np.ndarray), (
        "The covariance matrix of the observations should be a Numpy array."
    )
    assert len(sigma_yy.shape) == 2, (
        "The covariance matrix of the observations should be a 2D Numpy array."
    )
    assert np.shape(sigma_yy)[0] == np.shape(sigma_yy)[1], (
        "The covariance matrix of the observations should be a square matrix."
    )
    assert isinstance(sigma_ss, np.ndarray), (
        "The covariance matrix of the margins should be a Numpy array."
    )
    assert len(sigma_ss.shape) == 2, (
        "The covariance matrix of the margins should be a 2D Numpy array."
    )
    assert np.shape(sigma_ss)[0] == np.shape(sigma_ss)[1], (
        "The covariance matrix of the margins should be a square matrix."
    )
    assert isinstance(sigma_ys, np.ndarray), (
        "The covariance matrix of the observations and margins should be a Numpy array."
    )
    assert len(sigma_ys.shape) == 2, (
        "The covariance matrix of the observations and margins should be a 2D Numpy array."
    )
    assert np.shape(sigma_ys)[0] == np.shape(sigma_yy)[0], (
        "The covariance matrix of observations and margins should have the same number of rows as the covariance matrix of the observations."
    )
    assert np.shape(sigma_ys)[1] == np.shape(sigma_ss)[1], (
        "The covariance matrix of observations and margins should have the same number of columns as the covariance matrix of the margins."
    )


def check_covariance(
    sigma_yy: np.ndarray,
    sigma_ss: np.ndarray,
    sigma_ys: np.ndarray,
    rtol: float = 1e-05,
    atol: float = 1e-08,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Check if the covariance matrix is definite positive.

    If it is not, assumes independence of the variables
    and return the diagonal matrix of the variances.

    Parameters
    ----------
    sigma_yy : np.ndarray
        Covariance matrix of the observations
    sigma_ss : np.ndarray
        Covariance matrix of the margins
    sigma_ys : np.ndarray
        Covariance matrix of the observations and margins
    rtol : float
        Relative tolerance to check whether the margins are consistant. See numpy.allclose documentation for details.
    atol : float
        Absolute tolerance to check whether the margins are consistant. See numpy.allclose documentation for details.

    Returns
    -------
    sigma_yy : np.ndarray
        Covariance matrix of the observations
    sigma_ss : np.ndarray
        Covariance matrix of the margins
    sigma_ys : np.ndarray
        Covariance matrix of the observations and margins
    """
    check_format_covariance(sigma_yy, sigma_ss, sigma_ys)
    sigma = np.concatenate(
        (
            np.concatenate((sigma_yy, sigma_ys), axis=1),
            np.concatenate((np.transpose(sigma_ys), sigma_ss), axis=1),
        ),
        axis=0,
    )
    valid = True
    if np.allclose(np.transpose(sigma), sigma, rtol, atol):
        valid = False
    if np.any(np.linalg.eig(sigma)[0] < 0.0):
        valid = False
    if not valid:
        sigma_yy = np.diag(np.diag(sigma_yy))
        sigma_ss = np.diag(np.diag(sigma_ss))
        sigma_ys = np.zeros(sigma_ys.shape)
    return sigma_yy, sigma_ss, sigma_ys



"""Module to run the raking problems"""

import numpy as np
import pandas as pd

from raking.compute_constraints import (
    constraints_1D,
    constraints_2D,
    constraints_3D,
    constraints_USHD,
)
from raking.compute_covariance import compute_covariance_obs
from raking.compute_covariance import (
    compute_covariance_margins_1D,
    compute_covariance_margins_2D,
    compute_covariance_margins_3D,
    compute_covariance_margins_USHD,
)
from raking.compute_covariance import (
    compute_covariance_obs_margins_1D,
    compute_covariance_obs_margins_2D,
    compute_covariance_obs_margins_3D,
)
from raking.compute_covariance import check_covariance
from raking.formatting_methods import (
    format_data_1D,
    format_data_2D,
    format_data_3D,
    format_data_USHD,
)
from raking.raking_methods import (
    raking_chi2,
    raking_entropic,
    raking_general,
    raking_logit,
)
from raking.uncertainty_methods import compute_covariance, compute_gradient

pd.options.mode.chained_assignment = None


def run_raking(
    dim: int | str,
    df_obs: pd.DataFrame,
    df_margins: list,
    var_names: list | None,
    draws: str = "draws",
    cov_mat: bool = True,
    sigma_yy: np.ndarray = None,
    sigma_ss: np.ndarray = None,
    sigma_ys: np.ndarray = None,
    method: str = "chi2",
    alpha: float = 1,
    weights: str = None,
    lower: str = None,
    upper: str = None,
    rtol: float = 1e-05,
    atol: float = 1e-08,
    gamma0: float = 1.0,
    max_iter: int = 500,
) -> np.ndarray:
    """
    This function allows the user to run the raking problem.

    Parameters
    ----------
    dim : integer or string
        Dimension of the raking problem (1, 2, 3) or special case (USHD)
    df_obs : pd.DataFrame
        Observations data
    df_margins : list of pd.DataFrame
        list of data frames contatining the margins data
    var_names : list of strings
        Names of the variables over which we rake (e.g. cause, race, county). None if using special case.
    draws: string
        Name of the column that contains the samples.
    cov_mat : boolean
        If True, compute the covariance matrix of the raked values
    sigma_yy: np.ndarray
        Covariance matrix of the observations. We assume that there are sorted by var3, var2, var1.
        If None, the observation data frame must contain samples and we compte the sample covariance matrix.
    sigma_ss: np.ndarray
        Covariance matrix of the margins.
        If None, the margins data frames must contain samples and we compte the sample covariance matrix.
    sigma_ys: np.ndarray
        Covariance matrix of the observations and the margins.
        If None, the observations and margins data frames must contain samples and we compte the sample covariance matrix.
    method : string
        Name of the distance function used for the raking.
        Possible values are chi2, entropic, general, logit
    alpha : float
        Parameter of the distance function, alpha=1 is the chi2 distance, alpha=0 is the entropic distance
    weights : string
        Name of the column containing the raking weights
    lower : string
        Name of the column containing the lower boundaries (for logit raking)
    upper : string
        Name of the column containing the upper boundaries (for logit raking)
    rtol : float
        Relative tolerance to check whether the margins are consistant. See numpy.allclose documentation for details.
    atol : float
        Absolute tolerance to check whether the margins are consistant. See numpy.allclose documentation for details.
    gamma0 : float
        Initial value for line search
    max_iter : int
        Number of iterations for Newton's root finding method

    Returns
    -------
    df_obs : pd.DataFrame
        The initial observations data frame with an additional column for the raked values
    """
    assert isinstance(dim, int) or isinstance(dim, str), (
        "The dimension of the raking problem must be an integer or string."
    )
    assert dim in [
        1,
        2,
        3,
        "USHD",
    ], "The dimension of the raking problem must be 1, 2, 3 or USHD."
    assert isinstance(cov_mat, bool), (
        "cov_mat indicates whether we compute the covariance matrix, must be True or False."
    )
    if dim in [1, 2, 3]:
        assert isinstance(var_names, list), (
            "The variables over which we rake must be entered as a list."
        )
        assert dim == len(var_names), (
            "The number of variables over which we rake must be equal to the dimension of the problem."
        )
    else:
        var_names = ["cause", "race", "county"]
    assert isinstance(df_margins, list), (
        "The margins data frames must be entered as a list."
    )
    if dim in [1, 2, 3]:
        assert dim == len(df_margins), (
            "The number of margins data frames must be equal to the dimension of the problem."
        )
    else:
        assert len(df_margins) == 1, (
            "There should be only one margins data frame in the list."
        )
    assert isinstance(method, str), (
        "The name of the distance function used for the raking must be a string."
    )
    assert method in [
        "chi2",
        "entropic",
        "general",
        "logit",
    ], "The distance function must be chi2, entropic, general or logit."

    df_obs = df_obs.copy(deep=True)

    # Compute the covariance matrix
    if cov_mat:
        if dim == 1:
            (sigma_yy, sigma_ss, sigma_ys) = compute_covariance_1D(
                df_obs,
                df_margins,
                var_names,
                draws,
                sigma_yy,
                sigma_ss,
                sigma_ys,
            )
        elif dim == 2:
            (sigma_yy, sigma_ss, sigma_ys) = compute_covariance_2D(
                df_obs,
                df_margins,
                var_names,
                draws,
                sigma_yy,
                sigma_ss,
                sigma_ys,
            )
        elif dim == 3:
            (sigma_yy, sigma_ss, sigma_ys) = compute_covariance_3D(
                df_obs,
                df_margins,
                var_names,
                draws,
                sigma_yy,
                sigma_ss,
                sigma_ys,
            )
        elif dim == "USHD":
            (sigma_yy, sigma_ss, sigma_ys) = compute_covariance_USHD(
                df_obs,
                df_margins,
                var_names,
                draws,
                sigma_yy,
                sigma_ss,
                sigma_ys,
            )
        else:
            pass
        # Check if matrix is definite positive
        (sigma_yy, sigma_ss, sigma_ys) = check_covariance(
            sigma_yy, sigma_ss, sigma_ys, rtol, atol
        )

    # Compute the mean (if we have draws)
    if cov_mat:
        if dim in [1, 2, 3]:
            (df_obs, df_margins) = compute_mean(
                df_obs, df_margins, var_names, draws
            )
        else:
            (df_obs, df_margins) = compute_mean(
                df_obs, df_margins, ["race_county"], draws
            )

    # Get the input variables for the raking
    if dim == 1:
        (y, s, q, l, h, A) = run_raking_1D(
            df_obs, df_margins, var_names, weights, lower, upper, rtol, atol
        )
    elif dim == 2:
        (y, s, q, l, h, A) = run_raking_2D(
            df_obs, df_margins, var_names, weights, lower, upper, atol, rtol
        )
    elif dim == 3:
        (y, s, q, l, h, A) = run_raking_3D(
            df_obs, df_margins, var_names, weights, lower, upper, rtol, atol
        )
    elif dim == "USHD":
        (y, s, q, l, h, A) = run_raking_USHD(
            df_obs, df_margins, weights, lower, upper, rtol, atol
        )
    else:
        pass

    # Rake
    if method == "chi2":
        (beta, lambda_k) = raking_chi2(y, A, s, q)
    elif method == "entropic":
        (beta, lambda_k, iter_eps) = raking_entropic(
            y, A, s, q, gamma0, max_iter
        )
    elif method == "general":
        (beta, lambda_k, iter_eps) = raking_general(
            y, A, s, alpha, q, gamma0, max_iter
        )
    elif method == "logit":
        (beta, lambda_k, iter_eps) = raking_logit(
            y, A, s, l, h, q, gamma0, max_iter
        )
    else:
        pass

    # Create data frame for the raked values
    var_names.reverse()
    df_obs.sort_values(by=var_names, inplace=True)
    df_obs["raked_value"] = beta

    # Compute the covariance matrix of the raked values
    if cov_mat:
        (Dphi_y, Dphi_s) = compute_gradient(
            beta, lambda_k, y, A, method, alpha, l, h, q
        )
        sigma = compute_covariance(Dphi_y, Dphi_s, sigma_yy, sigma_ss, sigma_ys)
        df_obs["variance"] = np.diag(sigma)
    else:
        Dphi_y = None
        Dphi_s = None
        sigma = None
    return (df_obs, Dphi_y, Dphi_s, sigma)


def run_raking_1D(
    df_obs: pd.DataFrame,
    df_margins: list,
    var_names: list,
    weights: str = None,
    lower: str = None,
    upper: str = None,
    rtol: float = 1e-05,
    atol: float = 1e-08,
) -> tuple[
    np.ndarray,
    np.ndarray,
    np.ndarray | None,
    np.ndarray | None,
    np.ndarray | None,
    np.ndarray,
]:
    """
    This function prepares variables to run the raking problem in 1D.

    Parameters
    ----------
    df_obs : pd.DataFrame
        Observations data
    df_margins : list of pd.DataFrame
        list of data frames contatining the margins data
    var_names : list of strings
        Names of the variables over which we rake (e.g. cause, race, county)
    weights : string
        Name of the column containing the raking weights
    lower : string
        Name of the column containing the lower boundaries (for logit raking)
    upper : string
        Name of the column containing the upper boundaries (for logit raking)
    rtol : float
        Relative tolerance to check whether the margins are consistant. See numpy.allclose documentation for details.
    atol : float
        Absolute tolerance to check whether the margins are consistant. See numpy.allclose documentation for details.

    Returns
    -------
    y : np.ndarray
        Vector of observations
    s : np.ndarray
        Margins vector
    q : np.ndarray
        Vector of weights
    l : np.ndarray
        Lower bounds for the observations
    h : np.ndarray
        Upper bounds for the observations
    A : np.ndarray
        Constraints matrix
    """
    df_margins = df_margins[0]
    var_name = var_names[0]
    (y, s, I, q, l, h) = format_data_1D(
        df_obs, df_margins, var_name, weights, lower, upper
    )
    (A, s) = constraints_1D(s, I)
    return (y, s, q, l, h, A)


def run_raking_2D(
    df_obs: pd.DataFrame,
    df_margins: list,
    var_names: list,
    weights: str = None,
    lower: str = None,
    upper: str = None,
    rtol: float = 1e-05,
    atol: float = 1e-08,
) -> tuple[
    np.ndarray,
    np.ndarray,
    np.ndarray | None,
    np.ndarray | None,
    np.ndarray | None,
    np.ndarray,
]:
    """
    This function prepares variables to run the raking problem in 2D.

    Parameters
    ----------
    df_obs : pd.DataFrame
        Observations data
    df_margins : list of pd.DataFrame
        list of data frames contatining the margins data
    var_names : list of strings
        Names of the variables over which we rake (e.g. cause, race, county)
    weights : string
        Name of the column containing the raking weights
    lower : string
        Name of the column containing the lower boundaries (for logit raking)
    upper : string
        Name of the column containing the upper boundaries (for logit raking)
    rtol : float
        Relative tolerance to check whether the margins are consistant. See numpy.allclose documentation for details.
    atol : float
        Absolute tolerance to check whether the margins are consistant. See numpy.allclose documentation for details.

    Returns
    -------
    y : np.ndarray
        Vector of observations
    s : np.ndarray
        Margins vector
    q : np.ndarray
        Vector of weights
    l : np.ndarray
        Lower bounds for the observations
    h : np.ndarray
        Upper bounds for the observations
    A : np.ndarray
        Constraints matrix
    """
    df_margins_1 = df_margins[0]
    df_margins_2 = df_margins[1]
    (y, s1, s2, I, J, q, l, h) = format_data_2D(
        df_obs, df_margins_1, df_margins_2, var_names, weights, lower, upper
    )
    (A, s) = constraints_2D(s1, s2, I, J, rtol, atol)
    return (y, s, q, l, h, A)


def run_raking_3D(
    df_obs: pd.DataFrame,
    df_margins: list,
    var_names: list,
    weights: str = None,
    lower: str = None,
    upper: str = None,
    rtol: float = 1e-05,
    atol: float = 1e-08,
) -> tuple[
    np.ndarray,
    np.ndarray,
    np.ndarray | None,
    np.ndarray | None,
    np.ndarray | None,
    np.ndarray,
]:
    """
    This function prepares variables to run the raking problem in 3D.

    Parameters
    ----------
    df_obs : pd.DataFrame
        Observations data
    df_margins : list of pd.DataFrame
        list of data frames contatining the margins data
    var_names : list of strings
        Names of the variables over which we rake (e.g. cause, race, county)
    weights : string
        Name of the column containing the raking weights
    lower : string
        Name of the column containing the lower boundaries (for logit raking)
    upper : string
        Name of the column containing the upper boundaries (for logit raking)
    rtol : float
        Relative tolerance to check whether the margins are consistant. See numpy.allclose documentation for details.
    atol : float
        Absolute tolerance to check whether the margins are consistant. See numpy.allclose documentation for details.

    Returns
    -------
    y : np.ndarray
        Vector of observations
    s : np.ndarray
        Margins vector
    q : np.ndarray
        Vector of weights
    l : np.ndarray
        Lower bounds for the observations
    h : np.ndarray
        Upper bounds for the observations
    A : np.ndarray
        Constraints matrix
    """
    df_margins_1 = df_margins[0]
    df_margins_2 = df_margins[1]
    df_margins_3 = df_margins[2]
    (y, s1, s2, s3, I, J, K, q, l, h) = format_data_3D(
        df_obs,
        df_margins_1,
        df_margins_2,
        df_margins_3,
        var_names,
        weights,
        lower,
        upper,
    )
    (A, s) = constraints_3D(s1, s2, s3, I, J, K, rtol, atol)
    return (y, s, q, l, h, A)


def run_raking_USHD(
    df_obs: pd.DataFrame,
    df_margins: list,
    weights: str = None,
    lower: str = None,
    upper: str = None,
    rtol: float = 1e-05,
    atol: float = 1e-08,
) -> tuple[
    np.ndarray,
    np.ndarray,
    np.ndarray | None,
    np.ndarray | None,
    np.ndarray | None,
    np.ndarray,
]:
    """
    This function prepares variables to run the raking problem for the USHD case.

    Parameters
    ----------
    df_obs : pd.DataFrame
        Observations data
    df_margins : list of pd.DataFrame
        list of data frames contatining the margins data
    weights : string
        Name of the column containing the raking weights
    lower : string
        Name of the column containing the lower boundaries (for logit raking)
    upper : string
        Name of the column containing the upper boundaries (for logit raking)
    rtol : float
        Relative tolerance to check whether the margins are consistant. See numpy.allclose documentation for details.
    atol : float
        Absolute tolerance to check whether the margins are consistant. See numpy.allclose documentation for details.

    Returns
    -------
    y : np.ndarray
        Vector of observations
    s : np.ndarray
        Margins vector
    q : np.ndarray
        Vector of weights
    l : np.ndarray
        Lower bounds for the observations
    h : np.ndarray
        Upper bounds for the observations
    A : np.ndarray
        Constraints matrix
    """
    df_margins = df_margins[0]
    (y, s, I, J, K, q, l, h) = format_data_USHD(
        df_obs,
        df_margins,
        weights,
        lower,
        upper,
    )
    (A, s) = constraints_USHD(s, I, J, K, rtol, atol)
    return (y, s, q, l, h, A)


def compute_covariance_1D(
    df_obs: pd.DataFrame,
    df_margins: pd.DataFrame,
    var_names: list,
    draws: str,
    sigma_yy: np.ndarray = None,
    sigma_ss: np.ndarray = None,
    sigma_ys: np.ndarray = None,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Compute the covariance matrix of observations and margins in 1D.

    Parameters
    ----------
    df_obs : pd.DataFrame
        Observations data
    df_margins : list of pd.DataFrame
        list of data frames contatining the margins data
    var_names : list of strings
        Names of the variables over which we rake (e.g. cause, race, county)
    draws: string
        Name of the column that contains the samples.
    sigma_yy : np.ndarray
        Covariance matrix of the observations
    sigma_ss : np.ndarray
        Covariance matrix of the margins
    sigma_ys : np.ndarray
        Covariance matrix of the observations and margins

    Returns
    -------
    sigma_yy : np.ndarray
        Covariance matrix of the observations
    sigma_ss : np.ndarray
        Covariance matrix of the margins
    sigma_ys : np.ndarray
        Covariance matrix of the observations and margins
    """
    df_margins = df_margins[0]
    if sigma_yy is None:
        sigma_yy = compute_covariance_obs(df_obs, var_names, draws)
    if sigma_ss is None:
        sigma_ss = compute_covariance_margins_1D(df_margins, var_names, draws)
    if sigma_ys is None:
        sigma_ys = compute_covariance_obs_margins_1D(
            df_obs, df_margins, var_names, draws
        )
    return (sigma_yy, sigma_ss, sigma_ys)


def compute_covariance_2D(
    df_obs: pd.DataFrame,
    df_margins: pd.DataFrame,
    var_names: list,
    draws: str,
    sigma_yy: np.ndarray,
    sigma_ss: np.ndarray,
    sigma_ys: np.ndarray,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Compute the covariance matrix of observations and margins in 2D.

    Parameters
    ----------
    df_obs : pd.DataFrame
        Observations data
    df_margins : list of pd.DataFrame
        list of data frames contatining the margins data
    var_names : list of strings
        Names of the variables over which we rake (e.g. cause, race, county)
    draws: string
        Name of the column that contains the samples.
    sigma_yy : np.ndarray
        Covariance matrix of the observations
    sigma_ss : np.ndarray
        Covariance matrix of the margins
    sigma_ys : np.ndarray
        Covariance matrix of the observations and margins

    Returns
    -------
    sigma_yy : np.ndarray
        Covariance matrix of the observations
    sigma_ss : np.ndarray
        Covariance matrix of the margins
    sigma_ys : np.ndarray
        Covariance matrix of the observations and margins
    """
    df_margins_1 = df_margins[0]
    df_margins_2 = df_margins[1]
    if sigma_yy is None:
        sigma_yy = compute_covariance_obs(df_obs, var_names, draws)
    if sigma_ss is None:
        sigma_ss = compute_covariance_margins_2D(
            df_margins_1, df_margins_2, var_names, draws
        )
    if sigma_ys is None:
        sigma_ys = compute_covariance_obs_margins_2D(
            df_obs, df_margins_1, df_margins_2, var_names, draws
        )
    return (sigma_yy, sigma_ss, sigma_ys)


def compute_covariance_3D(
    df_obs: pd.DataFrame,
    df_margins: pd.DataFrame,
    var_names: list,
    draws: str,
    sigma_yy: np.ndarray,
    sigma_ss: np.ndarray,
    sigma_ys: np.ndarray,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Compute the covariance matrix of observations and margins in 3D.

    Parameters
    ----------
    df_obs : pd.DataFrame
        Observations data
    df_margins : list of pd.DataFrame
        list of data frames contatining the margins data
    var_names : list of strings
        Names of the variables over which we rake (e.g. cause, race, county)
    draws: string
        Name of the column that contains the samples.
    sigma_yy : np.ndarray
        Covariance matrix of the observations
    sigma_ss : np.ndarray
        Covariance matrix of the margins
    sigma_ys : np.ndarray
        Covariance matrix of the observations and margins

    Returns
    -------
    sigma_yy : np.ndarray
        Covariance matrix of the observations
    sigma_ss : np.ndarray
        Covariance matrix of the margins
    sigma_ys : np.ndarray
        Covariance matrix of the observations and margins
    """
    df_margins_1 = df_margins[0]
    df_margins_2 = df_margins[1]
    df_margins_3 = df_margins[2]
    if sigma_yy is None:
        sigma_yy = compute_covariance_obs(df_obs, var_names, draws)
    if sigma_ss is None:
        sigma_ss = compute_covariance_margins_3D(
            df_margins_1, df_margins_2, df_margins_3, var_names, draws
        )
    if sigma_ys is None:
        sigma_ys = compute_covariance_obs_margins_3D(
            df_obs, df_margins_1, df_margins_2, df_margins_3, var_names, draws
        )
    return (sigma_yy, sigma_ss, sigma_ys)


def compute_covariance_USHD(
    df_obs: pd.DataFrame,
    df_margins: pd.DataFrame,
    var_names: list,
    draws: str,
    sigma_yy: np.ndarray,
    sigma_ss: np.ndarray,
    sigma_ys: np.ndarray,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Compute the covariance matrix of observations and margins for the USHD case.

    Parameters
    ----------
    df_obs : pd.DataFrame
        Observations data
    df_margins : list of pd.DataFrame
        list of data frames contatining the margins data
    var_names : list of strings
        Names of the variables over which we rake (cause, race, county)
    draws: string
        Name of the column that contains the samples.
    sigma_yy : np.ndarray
        Covariance matrix of the observations
    sigma_ss : np.ndarray
        Covariance matrix of the margins
    sigma_ys : np.ndarray
        Covariance matrix of the observations and margins

    Returns
    -------
    sigma_yy : np.ndarray
        Covariance matrix of the observations
    sigma_ss : np.ndarray
        Covariance matrix of the margins
    sigma_ys : np.ndarray
        Covariance matrix of the observations and margins
    """
    df_margins = df_margins[0]
    if sigma_yy is None:
        sigma_yy = compute_covariance_obs(df_obs, var_names, draws)
    if sigma_ss is None:
        I = len(df_obs["cause"].unique()) - 1
        J = len(df_obs["race"].unique()) - 1
        K = len(df_obs["county"].unique())
        sigma_ss = compute_covariance_margins_USHD(
            df_margins, ["cause"], draws, I, J, K
        )
    if sigma_ys is None:
        sigma_ys = np.zeros((sigma_yy.shape[0], sigma_ss.shape[0]))
    return (sigma_yy, sigma_ss, sigma_ys)


def compute_mean(
    df_obs: pd.DataFrame, df_margins: list, var_names: list, draws: str
) -> tuple[pd.DataFrame, list]:
    """Compute the means of the values over all the samples.

    Parameters
    ----------
    df_obs : pd.DataFrame
        Observations data
    df_margins : list of pd.DataFrame
        list of data frames contatining the margins data
    var_names : list of strings
        Names of the variables over which we rake (e.g. cause, race, county)
    draws: string
        Name of the column that contains the samples.

    Returns
    -------
    df_obs_mean : pd.DataFrame
        Means of observations data
    df_margins_mean : list of pd.DataFrame
        list of data frames contatining the mans of the margins data
    """
    columns = df_obs.columns.drop([draws, "value"]).to_list()
    df_obs_mean = (
        df_obs.groupby(columns).mean().reset_index().drop(columns=[draws])
    )
    df_margins_mean = []
    for df_margin, var_name in zip(df_margins, var_names):
        value_name = "value_agg_over_" + var_name
        columns = df_margin.columns.drop([draws, value_name]).to_list()
        if len(columns) == 0:
            df_margin_mean = pd.DataFrame(
                {value_name: np.array([df_margin.mean()[value_name]])}
            )
        else:
            df_margin_mean = (
                df_margin.groupby(columns)
                .mean()
                .reset_index()
                .drop(columns=[draws])
            )
        df_margins_mean.append(df_margin_mean)
    return (df_obs_mean, df_margins_mean)



import pytest
import numpy as np
from raking.compute_constraints import (
    constraints_1D,
    constraints_2D,
    constraints_3D,
    constraints_USHD,
)
from raking.raking_methods import (
    raking_chi2,
    raking_entropic,
    raking_general,
    raking_logit,
)


def test_chi2_raking_1D():
    # Generate balanced vector
    I = 3
    rng = np.random.default_rng(0)
    beta = rng.uniform(low=2.0, high=3.0, size=I)
    s = np.sum(beta)
    # Add noise
    y = beta + rng.normal(0.0, 0.1, size=len(beta))
    # Generate the constraints
    (A, s) = constraints_1D(s, I)
    # Rake using chi2 distance
    (beta_star, lambda_star) = raking_chi2(y, A, s)
    # Verify that the constraint A beta_star = s is respected
    assert np.allclose(np.matmul(A, beta_star), s), (
        "For the raking in 1D with the chi2 distance, the constraint A beta_star = s is not respected."
    )


def test_chi2_raking_2D():
    # Generate balanced matrix
    I = 3
    J = 5
    rng = np.random.default_rng(0)
    beta = rng.uniform(low=2.0, high=3.0, size=(I, J))
    s1 = np.sum(beta, axis=0)
    s2 = np.sum(beta, axis=1)
    # Add noise
    y = beta + rng.normal(0.0, 0.1, size=beta.shape)
    y = y.flatten(order="F")
    # Generate the constraints
    (A, s) = constraints_2D(s1, s2, I, J)
    # Rake using chi2 distance
    (beta_star, lambda_star) = raking_chi2(y, A, s)
    # Verify that the constraint A beta_star = s is respected
    assert np.allclose(np.matmul(A, beta_star), s), (
        "For the raking in 2D with the chi2 distance, the constraint A beta_star = s is not respected."
    )


def test_chi2_raking_3D():
    # Generate balanced matrix
    I = 3
    J = 4
    K = 5
    rng = np.random.default_rng(0)
    beta = rng.uniform(low=2.0, high=3.0, size=(I, J, K))
    s1 = np.sum(beta, axis=0)
    s2 = np.sum(beta, axis=1)
    s3 = np.sum(beta, axis=2)
    # Add noise
    y = beta + rng.normal(0.0, 0.1, size=beta.shape)
    y = y.flatten(order="F")
    # Generate the constraints
    (A, s) = constraints_3D(s1, s2, s3, I, J, K)
    # Rake using chi2 distance
    (beta_star, lambda_star) = raking_chi2(y, A, s)
    # Verify that the constraint A beta_star = s is respected
    assert np.allclose(np.matmul(A, beta_star), s), (
        "For the raking in 3D with the chi2 distance, the constraint A beta_star = s is not respected."
    )


def test_chi2_raking_USHD():
    # Generate balanced array
    I = 3
    J = 4
    K = 5
    rng = np.random.default_rng(0)
    beta_ijk = rng.uniform(low=2.0, high=3.0, size=(I, J, K))
    beta_00k = np.sum(beta_ijk, axis=(0, 1))
    beta_i0k = np.sum(beta_ijk, axis=1)
    beta_0jk = np.sum(beta_ijk, axis=0)
    beta1 = np.concatenate(
        (beta_00k.reshape((1, 1, K)), beta_i0k.reshape(I, 1, K)), axis=0
    )
    beta2 = np.concatenate((beta_0jk.reshape((1, J, K)), beta_ijk), axis=0)
    beta = np.concatenate((beta1, beta2), axis=1)
    beta_i = np.sum(beta_ijk, axis=(1, 2))
    beta_0 = np.sum(beta_i)
    s_cause = np.array([beta_0] + beta_i.tolist())
    # Add noise
    y = beta + rng.normal(0.0, 0.1, size=beta.shape)
    y = y.flatten(order="F")
    # Generate the constraints
    (A, s) = constraints_USHD(s_cause, I, J, K)
    # Rake using chi2 distance
    (beta_star, lambda_star) = raking_chi2(y, A, s)
    # Verify that the constraint A beta_star = s is respected
    assert np.allclose(np.matmul(A, beta_star), s, atol=1.0e-5), (
        "For the USHD raking with the chi2 distance, the constraint A beta_star = s is not respected."
    )


def test_entropic_raking_1D():
    # Generate balanced vector
    I = 3
    rng = np.random.default_rng(0)
    beta = rng.uniform(low=2.0, high=3.0, size=I)
    s = np.sum(beta)
    # Add noise
    y = beta + rng.normal(0.0, 0.1, size=len(beta))
    # Generate the constraints
    (A, s) = constraints_1D(s, I)
    # Rake using entropic distance
    (beta_star, lambda_star, iter_eps) = raking_entropic(y, A, s)
    # Verify that the constraint A beta_star = s is respected
    assert np.allclose(np.matmul(A, beta_star), s), (
        "For the raking in 1D with the entropic distance, the constraint A beta_star = s is not respected."
    )


def test_entropic_raking_2D():
    # Generate balanced matrix
    I = 3
    J = 5
    rng = np.random.default_rng(0)
    beta = rng.uniform(low=2.0, high=3.0, size=(I, J))
    s1 = np.sum(beta, axis=0)
    s2 = np.sum(beta, axis=1)
    # Add noise
    y = beta + rng.normal(0.0, 0.1, size=beta.shape)
    y = y.flatten(order="F")
    # Generate the constraints
    (A, s) = constraints_2D(s1, s2, I, J)
    # Rake using entropic distance
    (beta_star, lambda_star, iter_eps) = raking_entropic(y, A, s)
    # Verify that the constraint A beta_star = s is respected
    assert np.allclose(np.matmul(A, beta_star), s), (
        "For the raking in 2D with the entropic distance, the constraint A beta_star = s is not respected."
    )


def test_entropic_raking_3D():
    # Generate balanced matrix
    I = 3
    J = 4
    K = 5
    rng = np.random.default_rng(0)
    beta = rng.uniform(low=2.0, high=3.0, size=(I, J, K))
    s1 = np.sum(beta, axis=0)
    s2 = np.sum(beta, axis=1)
    s3 = np.sum(beta, axis=2)
    # Add noise
    y = beta + rng.normal(0.0, 0.1, size=beta.shape)
    y = y.flatten(order="F")
    # Generate the constraints
    (A, s) = constraints_3D(s1, s2, s3, I, J, K)
    # Rake using entropic distance
    (beta_star, lambda_star, iter_eps) = raking_entropic(y, A, s)
    # Verify that the constraint A beta_star = s is respected
    assert np.allclose(np.matmul(A, beta_star), s), (
        "For the raking in 3D with the entropic distance, the constraint A beta_star = s is not respected."
    )


def test_entropic_raking_USHD():
    # Generate balanced array
    I = 3
    J = 4
    K = 5
    rng = np.random.default_rng(0)
    beta_ijk = rng.uniform(low=2.0, high=3.0, size=(I, J, K))
    beta_00k = np.sum(beta_ijk, axis=(0, 1))
    beta_i0k = np.sum(beta_ijk, axis=1)
    beta_0jk = np.sum(beta_ijk, axis=0)
    beta1 = np.concatenate(
        (beta_00k.reshape((1, 1, K)), beta_i0k.reshape(I, 1, K)), axis=0
    )
    beta2 = np.concatenate((beta_0jk.reshape((1, J, K)), beta_ijk), axis=0)
    beta = np.concatenate((beta1, beta2), axis=1)
    beta_i = np.sum(beta_ijk, axis=(1, 2))
    beta_0 = np.sum(beta_i)
    s_cause = np.array([beta_0] + beta_i.tolist())
    # Add noise
    y = beta + rng.normal(0.0, 0.1, size=beta.shape)
    y = y.flatten(order="F")
    # Generate the constraints
    (A, s) = constraints_USHD(s_cause, I, J, K)
    # Rake using entropic distance
    (beta_star, lambda_star, iter_eps) = raking_entropic(y, A, s)
    # Verify that the constraint A beta_star = s is respected
    assert np.allclose(np.matmul(A, beta_star), s, atol=1.0e-6), (
        "For the USHD raking with the entropic distance, the constraint A beta_star = s is not respected."
    )


def test_general_raking_1D():
    # Generate balanced vector
    I = 3
    rng = np.random.default_rng(0)
    beta = rng.uniform(low=2.0, high=3.0, size=I)
    s = np.sum(beta)
    # Add noise
    y = beta + rng.normal(0.0, 0.1, size=len(beta))
    # Generate the constraints
    (A, s) = constraints_1D(s, I)
    # Rake using general distance
    (beta_star, lambda_star, iter_eps) = raking_general(y, A, s, -2.0)
    # Verify that the constraint A beta_star = s is respected
    assert np.allclose(np.matmul(A, beta_star), s), (
        "For the raking in 1D with the general distance, the constraint A beta_star = s is not respected."
    )


def test_general_raking_2D():
    # Generate balanced matrix
    I = 3
    J = 5
    rng = np.random.default_rng(0)
    beta = rng.uniform(low=2.0, high=3.0, size=(I, J))
    s1 = np.sum(beta, axis=0)
    s2 = np.sum(beta, axis=1)
    # Add noise
    y = beta + rng.normal(0.0, 0.1, size=beta.shape)
    y = y.flatten(order="F")
    # Generate the constraints
    (A, s) = constraints_2D(s1, s2, I, J)
    # Rake using general distance
    (beta_star, lambda_star, iter_eps) = raking_general(y, A, s, -2.0)
    # Verify that the constraint A beta_star = s is respected
    assert np.allclose(np.matmul(A, beta_star), s), (
        "For the raking in 2D with the general distance, the constraint A beta_star = s is not respected."
    )


def test_general_raking_3D():
    # Generate balanced matrix
    I = 3
    J = 4
    K = 5
    rng = np.random.default_rng(0)
    beta = rng.uniform(low=2.0, high=3.0, size=(I, J, K))
    s1 = np.sum(beta, axis=0)
    s2 = np.sum(beta, axis=1)
    s3 = np.sum(beta, axis=2)
    # Add noise
    y = beta + rng.normal(0.0, 0.1, size=beta.shape)
    y = y.flatten(order="F")
    # Generate the constraints
    (A, s) = constraints_3D(s1, s2, s3, I, J, K)
    # Rake using general distance
    (beta_star, lambda_star, iter_eps) = raking_general(y, A, s, -2.0)
    # Verify that the constraint A beta_star = s is respected
    assert np.allclose(np.matmul(A, beta_star), s), (
        "For the raking in 3D with the general distance, the constraint A beta_star = s is not respected."
    )


def test_general_raking_USHD():
    # Generate balanced array
    I = 3
    J = 4
    K = 5
    rng = np.random.default_rng(0)
    beta_ijk = rng.uniform(low=2.0, high=3.0, size=(I, J, K))
    beta_00k = np.sum(beta_ijk, axis=(0, 1))
    beta_i0k = np.sum(beta_ijk, axis=1)
    beta_0jk = np.sum(beta_ijk, axis=0)
    beta1 = np.concatenate(
        (beta_00k.reshape((1, 1, K)), beta_i0k.reshape(I, 1, K)), axis=0
    )
    beta2 = np.concatenate((beta_0jk.reshape((1, J, K)), beta_ijk), axis=0)
    beta = np.concatenate((beta1, beta2), axis=1)
    beta_i = np.sum(beta_ijk, axis=(1, 2))
    beta_0 = np.sum(beta_i)
    s_cause = np.array([beta_0] + beta_i.tolist())
    # Add noise
    y = beta + rng.normal(0.0, 0.1, size=beta.shape)
    y = y.flatten(order="F")
    # Generate the constraints
    (A, s) = constraints_USHD(s_cause, I, J, K)
    # Rake using general distance
    (beta_star, lambda_star, iter_eps) = raking_general(y, A, s, -2.0)
    # Verify that the constraint A beta_star = s is respected
    assert np.allclose(np.matmul(A, beta_star), s, atol=1.0e-6), (
        "For the USHD raking with the general distance, the constraint A beta_star = s is not respected."
    )


def test_logit_raking_1D():
    # Generate balanced vector
    I = 3
    rng = np.random.default_rng(0)
    beta = rng.uniform(low=2.0, high=3.0, size=I)
    s = np.sum(beta)
    # Add noise
    y = beta + rng.normal(0.0, 0.1, size=len(beta))
    l = np.repeat(np.min(y), len(y))
    h = np.repeat(np.max(y), len(y))
    # Generate the constraints
    (A, s) = constraints_1D(s, I)
    # Rake using logit distance
    (beta_star, lambda_star, iter_eps) = raking_logit(y, A, s, l, h)
    # Verify that the constraint A beta_star = s is respected
    assert np.allclose(np.matmul(A, beta_star), s), (
        "For the raking in 1D with the logit distance, the constraint A beta_star = s is not respected."
    )
    # Verify that the lower bound is respected
    assert np.all(beta_star - l > -1.0e-5), (
        "For the raking in 1D with the logit distance, some raked values are lower than the lower bound."
    )
    # Verify that the upper bound is respected
    assert np.all(h - beta_star > -1.0e-5), (
        "For the raking in 1D with the logit distance, some raked values are higher than the upper bound."
    )


def test_logit_raking_2D():
    # Generate balanced matrix
    I = 3
    J = 5
    rng = np.random.default_rng(0)
    beta = rng.uniform(low=2.0, high=3.0, size=(I, J))
    s1 = np.sum(beta, axis=0)
    s2 = np.sum(beta, axis=1)
    # Add noise
    y = beta + rng.normal(0.0, 0.1, size=beta.shape)
    y = y.flatten(order="F")
    l = np.repeat(np.min(y), len(y))
    h = np.repeat(np.max(y), len(y))
    # Generate the constraints
    (A, s) = constraints_2D(s1, s2, I, J)
    # Rake using logit distance
    (beta_star, lambda_star, iter_eps) = raking_logit(y, A, s, l, h)
    # Verify that the constraint A beta_star = s is respected
    assert np.allclose(np.matmul(A, beta_star), s), (
        "For the raking in 2D with the logit distance, the constraint A beta_star = s is not respected."
    )
    # Verify that the lower bound is respected
    assert np.all(beta_star - l > -1.0e-5), (
        "For the raking in 2D with the logit distance, some raked values are lower than the lower bound."
    )
    # Verify that the upper bound is respected
    assert np.all(h - beta_star > -1.0e-5), (
        "For the raking in 2D with the logit distance, some raked values are higher than the upper bound."
    )


def test_logit_raking_3D():
    # Generate balanced matrix
    I = 3
    J = 4
    K = 5
    rng = np.random.default_rng(0)
    beta = rng.uniform(low=2.0, high=3.0, size=(I, J, K))
    s1 = np.sum(beta, axis=0)
    s2 = np.sum(beta, axis=1)
    s3 = np.sum(beta, axis=2)
    # Add noise
    y = beta + rng.normal(0.0, 0.1, size=beta.shape)
    y = y.flatten(order="F")
    l = np.repeat(np.min(y), len(y))
    h = np.repeat(np.max(y), len(y))
    # Generate the constraints
    (A, s) = constraints_3D(s1, s2, s3, I, J, K)
    # Rake using logitdistance
    (beta_star, lambda_star, iter_eps) = raking_logit(y, A, s, l, h)
    # Verify that the constraint A beta_star = s is respected
    assert np.allclose(np.matmul(A, beta_star), s), (
        "For the raking in 3D with the logit distance, the constraint A beta_star = s is not respected."
    )
    # Verify that the lower bound is respected
    assert np.all(beta_star - l > -1.0e-5), (
        "For the raking in 3D with the logit distance, some raked values are lower than the lower bound."
    )
    # Verify that the upper bound is respected
    assert np.all(h - beta_star > -1.0e-5), (
        "For the raking in 3D with the logit distance, some raked values are higher than the upper bound."
    )


def test_logit_raking_USHD():
    # Generate balanced array
    I = 3
    J = 4
    K = 5
    rng = np.random.default_rng(0)
    beta_ijk = rng.uniform(low=2.0, high=3.0, size=(I, J, K))
    beta_00k = np.sum(beta_ijk, axis=(0, 1))
    beta_i0k = np.sum(beta_ijk, axis=1)
    beta_0jk = np.sum(beta_ijk, axis=0)
    beta1 = np.concatenate(
        (beta_00k.reshape((1, 1, K)), beta_i0k.reshape(I, 1, K)), axis=0
    )
    beta2 = np.concatenate((beta_0jk.reshape((1, J, K)), beta_ijk), axis=0)
    beta = np.concatenate((beta1, beta2), axis=1)
    beta_i = np.sum(beta_ijk, axis=(1, 2))
    beta_0 = np.sum(beta_i)
    s_cause = np.array([beta_0] + beta_i.tolist())
    # Add noise
    y = beta + rng.normal(0.0, 0.1, size=beta.shape)
    y = y.flatten(order="F")
    l = np.repeat(np.min(y), len(y))
    h = np.repeat(np.max(y), len(y))
    # Generate the constraints
    (A, s) = constraints_USHD(s_cause, I, J, K)
    # Rake using logit distance
    (beta_star, lambda_star, iter_eps) = raking_logit(y, A, s, l, h)
    # Verify that the constraint A beta_star = s is respected
    assert np.allclose(np.matmul(A, beta_star), s, atol=1.0e-6), (
        "For the USHD raking with the logit distance, the constraint A beta_star = s is not respected."
    )
    # Verify that the lower bound is respected
    assert np.all(beta_star - l > -1.0e-5), (
        "For the USHD raking with the logit distance, some raked values are lower than the lower bound."
    )
    # Verify that the upper bound is respected
    assert np.all(h - beta_star > -1.0e-5), (
        "For the USHD raking with the logit distance, some raked values are higher than the upper bound."
    )
